@@ -29,10 +29,14 @@ typedef struct {
 //-----------------------------------------------------------------------------
 // constants for the OCI attributes
 //-----------------------------------------------------------------------------
-#ifdef ORACLE_10G
+static ub4 gc_ClientIdentifierAttribute = OCI_ATTR_CLIENT_IDENTIFIER;
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 1)
 static ub4 gc_ModuleAttribute = OCI_ATTR_MODULE;
 static ub4 gc_ActionAttribute = OCI_ATTR_ACTION;
 static ub4 gc_ClientInfoAttribute = OCI_ATTR_CLIENT_INFO;
+#endif
+
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
 static ub4 gc_CurrentSchemaAttribute = OCI_ATTR_CURRENT_SCHEMA;
 #endif
 
@@ -54,22 +58,22 @@ static PyObject *Connection_Cancel(udt_Connection*, PyObject*);
 static PyObject *Connection_RegisterCallback(udt_Connection*, PyObject*);
 static PyObject *Connection_UnregisterCallback(udt_Connection*, PyObject*);
 static PyObject *Connection_GetVersion(udt_Connection*, void*);
+static PyObject *Connection_GetEncoding(udt_Connection*, void*);
+static PyObject *Connection_GetNationalEncoding(udt_Connection*, void*);
 static PyObject *Connection_GetMaxBytesPerCharacter(udt_Connection*, void*);
 static PyObject *Connection_ContextManagerEnter(udt_Connection*, PyObject*);
 static PyObject *Connection_ContextManagerExit(udt_Connection*, PyObject*);
 static PyObject *Connection_ChangePasswordExternal(udt_Connection*, PyObject*);
-#ifndef WITH_UNICODE
-static PyObject *Connection_GetEncoding(udt_Connection*, void*);
-static PyObject *Connection_GetNationalEncoding(udt_Connection*, void*);
-#endif
 static PyObject *Connection_GetStmtCacheSize(udt_Connection*, void*);
 static int Connection_SetStmtCacheSize(udt_Connection*, PyObject*, void*);
-#ifdef ORACLE_10G
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
 static PyObject *Connection_GetOCIAttr(udt_Connection*, ub4*);
-static int Connection_SetOCIAttr(udt_Connection*, PyObject*, ub4*);
 #endif
-#ifdef ORACLE_10GR2
+static int Connection_SetOCIAttr(udt_Connection*, PyObject*, ub4*);
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
+#if !defined(AIX5) || defined(ORACLE_11g)
 static PyObject *Connection_Ping(udt_Connection*, PyObject*);
+#endif
 static PyObject *Connection_Shutdown(udt_Connection*, PyObject*, PyObject*);
 static PyObject *Connection_Startup(udt_Connection*, PyObject*, PyObject*);
 static PyObject *Connection_Subscribe(udt_Connection*, PyObject*, PyObject*);
@@ -91,8 +95,10 @@ static PyMethodDef g_ConnectionMethods[] = {
     { "unregister", (PyCFunction) Connection_UnregisterCallback, METH_VARARGS },
     { "__enter__", (PyCFunction) Connection_ContextManagerEnter, METH_NOARGS },
     { "__exit__", (PyCFunction) Connection_ContextManagerExit, METH_VARARGS },
-#ifdef ORACLE_10GR2
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
+#if !defined(AIX5) || ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
     { "ping", (PyCFunction) Connection_Ping, METH_NOARGS },
+#endif
     { "shutdown", (PyCFunction) Connection_Shutdown,
             METH_VARARGS | METH_KEYWORDS},
     { "startup", (PyCFunction) Connection_Startup,
@@ -127,20 +133,22 @@ static PyMemberDef g_ConnectionMembers[] = {
 // declaration of calculated members for Python type "Connection"
 //-----------------------------------------------------------------------------
 static PyGetSetDef g_ConnectionCalcMembers[] = {
-#ifndef WITH_UNICODE
+    { "version", (getter) Connection_GetVersion, 0, 0, 0 },
     { "encoding", (getter) Connection_GetEncoding, 0, 0, 0 },
     { "nencoding", (getter) Connection_GetNationalEncoding, 0, 0, 0 },
-#endif
-    { "version", (getter) Connection_GetVersion, 0, 0, 0 },
     { "maxBytesPerCharacter", (getter) Connection_GetMaxBytesPerCharacter,
             0, 0, 0 },
     { "stmtcachesize", (getter) Connection_GetStmtCacheSize,
             (setter) Connection_SetStmtCacheSize, 0, 0 },
-#ifdef ORACLE_10G
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 1)
     { "module", 0, (setter) Connection_SetOCIAttr, 0, &gc_ModuleAttribute },
     { "action", 0, (setter) Connection_SetOCIAttr, 0, &gc_ActionAttribute },
     { "clientinfo", 0, (setter) Connection_SetOCIAttr, 0,
             &gc_ClientInfoAttribute },
+#endif
+    { "client_identifier", 0, (setter) Connection_SetOCIAttr, 0,
+            &gc_ClientIdentifierAttribute },
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
     { "current_schema", (getter) Connection_GetOCIAttr,
             (setter) Connection_SetOCIAttr, 0, &gc_CurrentSchemaAttribute },
 #endif
@@ -226,7 +234,7 @@ static int Connection_GetConnection(
 {
     int externalCredentials, proxyCredentials;
     udt_Environment *environment;
-    udt_StringBuffer buffer;
+    udt_Buffer buffer;
     OCIAuthInfo *authInfo;
     PyObject *dbNameObj;
     boolean found;
@@ -265,7 +273,8 @@ static int Connection_GetConnection(
 
         // set the user name, if applicable
         externalCredentials = 1;
-        if (StringBuffer_Fill(&buffer, self->username) < 0)
+        if (cxBuffer_FromObject(&buffer, self->username,
+                self->environment->encoding) < 0)
             return -1;
         if (buffer.size > 0) {
             externalCredentials = 0;
@@ -274,14 +283,15 @@ static int Connection_GetConnection(
                     environment->errorHandle);
             if (Environment_CheckForError(environment, status,
                     "Connection_GetConnection(): set user name") < 0) {
-                StringBuffer_Clear(&buffer);
+                cxBuffer_Clear(&buffer);
                 return -1;
             }
         }
-        StringBuffer_Clear(&buffer);
+        cxBuffer_Clear(&buffer);
 
         // set the password, if applicable
-        if (StringBuffer_Fill(&buffer, self->password) < 0)
+        if (cxBuffer_FromObject(&buffer, self->password,
+                self->environment->encoding) < 0)
             return -1;
         if (buffer.size > 0) {
             externalCredentials = 0;
@@ -290,19 +300,20 @@ static int Connection_GetConnection(
                     environment->errorHandle);
             if (Environment_CheckForError(environment, status,
                     "Connection_GetConnection(): set password") < 0) {
-                StringBuffer_Clear(&buffer);
+                cxBuffer_Clear(&buffer);
                 return -1;
             }
         }
-        StringBuffer_Clear(&buffer);
+        cxBuffer_Clear(&buffer);
 
         // if no user name or password are set, using external credentials
         if (!pool && externalCredentials)
             mode |= OCI_SESSGET_CREDEXT;
 
-#ifdef ORACLE_11G
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
         // set the connection class, if applicable
-        if (StringBuffer_Fill(&buffer, cclassObj) < 0)
+        if (cxBuffer_FromObject(&buffer, cclassObj,
+                self->environment->encoding) < 0)
             return -1;
         if (buffer.size > 0) {
             status = OCIAttrSet(authInfo, OCI_HTYPE_AUTHINFO,
@@ -310,11 +321,11 @@ static int Connection_GetConnection(
                     environment->errorHandle);
             if (Environment_CheckForError(environment, status,
                     "Connection_GetConnection(): set connection class") < 0) {
-                StringBuffer_Clear(&buffer);
+                cxBuffer_Clear(&buffer);
                 return -1;
             }
         }
-        StringBuffer_Clear(&buffer);
+        cxBuffer_Clear(&buffer);
 
         // set the purity, if applicable
         if (purity != OCI_ATTR_PURITY_DEFAULT) {
@@ -329,14 +340,15 @@ static int Connection_GetConnection(
     }
 
     // acquire the new session
-    if (StringBuffer_Fill(&buffer, dbNameObj) < 0)
+    if (cxBuffer_FromObject(&buffer, dbNameObj,
+            self->environment->encoding) < 0)
         return -1;
     Py_BEGIN_ALLOW_THREADS
     status = OCISessionGet(environment->handle, environment->errorHandle,
             &self->handle, authInfo, (text*) buffer.ptr, buffer.size, NULL, 0,
             NULL, NULL, &found, mode);
     Py_END_ALLOW_THREADS
-    StringBuffer_Clear(&buffer);
+    cxBuffer_Clear(&buffer);
     if (Environment_CheckForError(environment, status,
             "Connection_GetConnection(): get connection") < 0)
         return -1;
@@ -364,7 +376,7 @@ static int Connection_GetConnection(
 }
 
 
-#ifdef ORACLE_10G
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
 //-----------------------------------------------------------------------------
 // Connection_GetOCIAttr()
 //   Get the value of the OCI attribute.
@@ -374,7 +386,8 @@ static PyObject *Connection_GetOCIAttr(
     ub4 *attribute)                     // OCI attribute type
 {
     OCISession *sessionHandle;
-    udt_StringBuffer buffer;
+    udt_Buffer buffer;
+    ub4 bufferSize;
     sword status;
 
     // make sure connection is connected
@@ -386,19 +399,22 @@ static PyObject *Connection_GetOCIAttr(
             (dvoid**) &sessionHandle, 0, OCI_ATTR_SESSION,
             self->environment->errorHandle);
     if (Environment_CheckForError(self->environment, status,
-            "Connection_SetOCIAttr(): determine session handle") < 0)
+            "Connection_GetOCIAttr(): determine session handle") < 0)
         return NULL;
 
     // get the value from the OCI
     status = OCIAttrGet(sessionHandle, OCI_HTYPE_SESSION,
-            (text**) &buffer.ptr, (ub4*) &buffer.size, *attribute,
+            (text**) &buffer.ptr, &bufferSize, *attribute,
             self->environment->errorHandle);
     if (Environment_CheckForError(self->environment, status,
             "Connection_GetOCIAttr()") < 0)
         return NULL;
 
-    return cxString_FromEncodedString(buffer.ptr, buffer.size);
+    buffer.size = bufferSize;
+    return cxString_FromEncodedString(buffer.ptr, buffer.size,
+            self->environment->encoding);
 }
+#endif
 
 
 //-----------------------------------------------------------------------------
@@ -411,7 +427,7 @@ static int Connection_SetOCIAttr(
     ub4 *attribute)                     // OCI attribute type
 {
     OCISession *sessionHandle;
-    udt_StringBuffer buffer;
+    udt_Buffer buffer;
     sword status;
 
     // verify arguments
@@ -433,17 +449,16 @@ static int Connection_SetOCIAttr(
         return -1;
 
     // set the value in the OCI
-    if (StringBuffer_Fill(&buffer, value))
+    if (cxBuffer_FromObject(&buffer, value, self->environment->encoding))
         return -1;
     status = OCIAttrSet(sessionHandle, OCI_HTYPE_SESSION, (text*) buffer.ptr,
             buffer.size, *attribute, self->environment->errorHandle);
-    StringBuffer_Clear(&buffer);
+    cxBuffer_Clear(&buffer);
     if (Environment_CheckForError(self->environment, status,
-            "Connection_SetOCIAttr()") < 0)
+            "Connection_SetOCIAttr(): set value") < 0)
         return -1;
     return 0;
 }
-#endif
 
 
 //-----------------------------------------------------------------------------
@@ -507,18 +522,21 @@ static int Connection_ChangePassword(
     PyObject *oldPasswordObj,           // old password
     PyObject *newPasswordObj)           // new password
 {
-    udt_StringBuffer usernameBuffer, oldPasswordBuffer, newPasswordBuffer;
+    udt_Buffer usernameBuffer, oldPasswordBuffer, newPasswordBuffer;
     sword status;
 
-    if (StringBuffer_Fill(&usernameBuffer, self->username) < 0)
+    if (cxBuffer_FromObject(&usernameBuffer, self->username,
+            self->environment->encoding) < 0)
         return -1;
-    if (StringBuffer_Fill(&oldPasswordBuffer, oldPasswordObj) < 0) {
-        StringBuffer_Clear(&usernameBuffer);
+    if (cxBuffer_FromObject(&oldPasswordBuffer, oldPasswordObj,
+            self->environment->encoding) < 0) {
+        cxBuffer_Clear(&usernameBuffer);
         return -1;
     }
-    if (StringBuffer_Fill(&newPasswordBuffer, newPasswordObj) < 0) {
-        StringBuffer_Clear(&usernameBuffer);
-        StringBuffer_Clear(&oldPasswordBuffer);
+    if (cxBuffer_FromObject(&newPasswordBuffer, newPasswordObj,
+            self->environment->encoding) < 0) {
+        cxBuffer_Clear(&usernameBuffer);
+        cxBuffer_Clear(&oldPasswordBuffer);
         return -1;
     }
 
@@ -530,9 +548,9 @@ static int Connection_ChangePassword(
             (text*) newPasswordBuffer.ptr, newPasswordBuffer.size,
             OCI_AUTH);
     Py_END_ALLOW_THREADS
-    StringBuffer_Clear(&usernameBuffer);
-    StringBuffer_Clear(&oldPasswordBuffer);
-    StringBuffer_Clear(&newPasswordBuffer);
+    cxBuffer_Clear(&usernameBuffer);
+    cxBuffer_Clear(&oldPasswordBuffer);
+    cxBuffer_Clear(&newPasswordBuffer);
     if (Environment_CheckForError(self->environment, status,
             "Connection_ChangePassword(): change password") < 0)
         return -1;
@@ -575,10 +593,13 @@ static int Connection_Connect(
     udt_Connection *self,               // connection
     ub4 mode,                           // mode to connect as
     int twophase,                       // allow two phase commit?
-    PyObject *newPasswordObj)           // new password (if desired)
+    PyObject *newPasswordObj,           // new password (if desired)
+    PyObject *moduleObj,                // session "module" value
+    PyObject *actionObj,                // session "action" value
+    PyObject *clientinfoObj)            // session "clientinfo" value
 {
     ub4 credentialType = OCI_CRED_EXT;
-    udt_StringBuffer buffer;
+    udt_Buffer buffer;
     sword status;
 
     // allocate the server handle
@@ -589,14 +610,15 @@ static int Connection_Connect(
         return -1;
 
     // attach to the server
-    if (StringBuffer_Fill(&buffer, self->dsn) < 0)
+    if (cxBuffer_FromObject(&buffer, self->dsn,
+            self->environment->encoding) < 0)
         return -1;
     Py_BEGIN_ALLOW_THREADS
     status = OCIServerAttach(self->serverHandle,
             self->environment->errorHandle, (text*) buffer.ptr, buffer.size,
             OCI_DEFAULT);
     Py_END_ALLOW_THREADS
-    StringBuffer_Clear(&buffer);
+    cxBuffer_Clear(&buffer);
     if (Environment_CheckForError(self->environment, status,
             "Connection_Connect(): server attach") < 0)
         return -1;
@@ -640,7 +662,8 @@ static int Connection_Connect(
         return -1;
 
     // set user name in session handle
-    if (StringBuffer_Fill(&buffer, self->username) < 0)
+    if (cxBuffer_FromObject(&buffer, self->username,
+            self->environment->encoding) < 0)
         return -1;
     if (buffer.size > 0) {
         credentialType = OCI_CRED_RDBMS;
@@ -649,14 +672,15 @@ static int Connection_Connect(
                 self->environment->errorHandle);
         if (Environment_CheckForError(self->environment, status,
                 "Connection_Connect(): set user name") < 0) {
-            StringBuffer_Clear(&buffer);
+            cxBuffer_Clear(&buffer);
             return -1;
         }
     }
-    StringBuffer_Clear(&buffer);
+    cxBuffer_Clear(&buffer);
 
     // set password in session handle
-    if (StringBuffer_Fill(&buffer, self->password) < 0)
+    if (cxBuffer_FromObject(&buffer, self->password,
+            self->environment->encoding) < 0)
         return -1;
     if (buffer.size > 0) {
         credentialType = OCI_CRED_RDBMS;
@@ -665,11 +689,11 @@ static int Connection_Connect(
                 self->environment->errorHandle);
         if (Environment_CheckForError(self->environment, status,
                 "Connection_Connect(): set password") < 0) {
-            StringBuffer_Clear(&buffer);
+            cxBuffer_Clear(&buffer);
             return -1;
         }
     }
-    StringBuffer_Clear(&buffer);
+    cxBuffer_Clear(&buffer);
 
 #ifdef OCI_ATTR_DRIVER_NAME
     status = OCIAttrSet(self->sessionHandle, OCI_HTYPE_SESSION,
@@ -689,6 +713,47 @@ static int Connection_Connect(
             "Connection_Connect(): set session handle") < 0)
         return -1;
 
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 1)
+    if (moduleObj) {
+        if (cxBuffer_FromObject(&buffer, moduleObj,
+                self->environment->encoding))
+            return -1;
+        status = OCIAttrSet(self->sessionHandle, OCI_HTYPE_SESSION,
+                (text*) buffer.ptr, buffer.size, OCI_ATTR_MODULE,
+                self->environment->errorHandle);
+        cxBuffer_Clear(&buffer);
+        if (Environment_CheckForError(self->environment, status,
+                "Connection_Connect(): set module") < 0)
+            return -1;
+    }
+
+    if (actionObj) {
+        if (cxBuffer_FromObject(&buffer, actionObj,
+                self->environment->encoding))
+            return -1;
+        status = OCIAttrSet(self->sessionHandle, OCI_HTYPE_SESSION,
+                (text*) buffer.ptr, buffer.size, OCI_ATTR_ACTION,
+                self->environment->errorHandle);
+        cxBuffer_Clear(&buffer);
+        if (Environment_CheckForError(self->environment, status,
+                "Connection_Connect(): set action") < 0)
+            return -1;
+    }
+
+    if (clientinfoObj) {
+        if (cxBuffer_FromObject(&buffer, clientinfoObj,
+                self->environment->encoding))
+            return -1;
+        status = OCIAttrSet(self->sessionHandle, OCI_HTYPE_SESSION,
+                (text*) buffer.ptr, buffer.size, OCI_ATTR_CLIENT_INFO,
+                self->environment->errorHandle);
+        cxBuffer_Clear(&buffer);
+        if (Environment_CheckForError(self->environment, status,
+                "Connection_Connect(): set clientinfo") < 0)
+            return -1;
+    }
+#endif
+
     // if a new password has been specified, change it which will also
     // establish the session
     if (newPasswordObj)
@@ -711,7 +776,7 @@ static int Connection_Connect(
 
 #include "Cursor.c"
 #include "Callback.c"
-#ifdef ORACLE_10GR2
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
 #include "Subscription.c"
 #endif
 
@@ -788,7 +853,9 @@ static int Connection_Init(
 {
     PyObject *threadedObj, *twophaseObj, *eventsObj, *newPasswordObj;
     PyObject *usernameObj, *passwordObj, *dsnObj, *cclassObj;
+    PyObject *moduleObj, *actionObj, *clientinfoObj;
     int threaded, twophase, events;
+    char *encoding, *nencoding;
     ub4 connectMode, purity;
     udt_SessionPool *pool;
     OCISvcCtx *handle;
@@ -796,7 +863,8 @@ static int Connection_Init(
     // define keyword arguments
     static char *keywordList[] = { "user", "password", "dsn", "mode",
             "handle", "pool", "threaded", "twophase", "events", "cclass",
-            "purity", "newpassword", NULL };
+            "purity", "newpassword", "encoding", "nencoding", "module",
+            "action", "clientinfo", NULL };
 
     // parse arguments
     pool = NULL;
@@ -804,16 +872,18 @@ static int Connection_Init(
     connectMode = OCI_DEFAULT;
     usernameObj = passwordObj = dsnObj = cclassObj = NULL;
     threadedObj = twophaseObj = eventsObj = newPasswordObj = NULL;
+    moduleObj = actionObj = clientinfoObj = NULL;
     threaded = twophase = events = purity = 0;
+    encoding = nencoding = NULL;
 #ifdef ORACLE_11G
     purity = OCI_ATTR_PURITY_DEFAULT;
 #endif
-    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "|O!O!O!iiO!OOOO!iO!",
-            keywordList, cxString_Type, &usernameObj, cxString_Type,
-            &passwordObj, cxString_Type, &dsnObj, &connectMode, &handle,
-            &g_SessionPoolType, &pool, &threadedObj, &twophaseObj, &eventsObj,
-            cxString_Type, &cclassObj, &purity, cxString_Type,
-            &newPasswordObj))
+    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs,
+            "|OOOiiO!OOOOiOssOOO", keywordList, &usernameObj, &passwordObj,
+            &dsnObj, &connectMode, &handle, &g_SessionPoolType, &pool,
+            &threadedObj, &twophaseObj, &eventsObj, &cclassObj, &purity,
+            &newPasswordObj, &encoding, &nencoding, &moduleObj, &actionObj,
+            &clientinfoObj))
         return -1;
     if (threadedObj) {
         threaded = PyObject_IsTrue(threadedObj);
@@ -834,7 +904,8 @@ static int Connection_Init(
     // set up the environment
     if (pool)
         self->environment = Environment_Clone(pool->environment);
-    else self->environment = Environment_NewFromScratch(threaded, events);
+    else self->environment = Environment_NewFromScratch(threaded, events,
+            encoding, nencoding);
     if (!self->environment)
         return -1;
 
@@ -857,7 +928,8 @@ static int Connection_Init(
         return Connection_Attach(self, handle);
     if (pool || cclassObj)
         return Connection_GetConnection(self, pool, cclassObj, purity);
-    return Connection_Connect(self, connectMode, twophase, newPasswordObj);
+    return Connection_Connect(self, connectMode, twophase, newPasswordObj,
+            moduleObj, actionObj, clientinfoObj);
 }
 
 
@@ -941,70 +1013,6 @@ static PyObject *Connection_Repr(
 }
 
 
-#ifndef WITH_UNICODE
-//-----------------------------------------------------------------------------
-// Connection_GetCharacterSetName()
-//   Retrieve the IANA character set name for the attribute.
-//-----------------------------------------------------------------------------
-static PyObject *Connection_GetCharacterSetName(
-    udt_Connection *self,               // connection object
-    ub2 attribute)                      // attribute to fetch
-{
-    char charsetName[OCI_NLS_MAXBUFSZ], ianaCharsetName[OCI_NLS_MAXBUFSZ];
-    ub2 charsetId;
-    sword status;
-
-    // get character set id
-    status = OCIAttrGet(self->environment->handle, OCI_HTYPE_ENV, &charsetId,
-            NULL, attribute, self->environment->errorHandle);
-    if (Environment_CheckForError(self->environment, status,
-            "Connection_GetCharacterSetName(): get charset id") < 0)
-        return NULL;
-
-    // get character set name
-    status = OCINlsCharSetIdToName(self->environment->handle,
-            (text*) charsetName, OCI_NLS_MAXBUFSZ, charsetId);
-    if (Environment_CheckForError(self->environment, status,
-            "Connection_GetCharacterSetName(): get Oracle charset name") < 0)
-        return NULL;
-
-    // get IANA character set name
-    status = OCINlsNameMap(self->environment->handle,
-            (oratext*) ianaCharsetName, OCI_NLS_MAXBUFSZ,
-            (oratext*) charsetName, OCI_NLS_CS_ORA_TO_IANA);
-    if (Environment_CheckForError(self->environment, status,
-            "Connection_GetCharacterSetName(): translate NLS charset") < 0)
-        return NULL;
-
-    return PyBytes_FromString(ianaCharsetName);
-}
-
-
-//-----------------------------------------------------------------------------
-// Connection_GetEncoding()
-//   Retrieve the IANA encoding used by the client.
-//-----------------------------------------------------------------------------
-static PyObject *Connection_GetEncoding(
-    udt_Connection *self,               // connection object
-    void *arg)                          // optional argument (ignored)
-{
-    return Connection_GetCharacterSetName(self, OCI_ATTR_ENV_CHARSET_ID);
-}
-
-
-//-----------------------------------------------------------------------------
-// Connection_GetNationalEncoding()
-//   Retrieve the IANA national encoding used by the client.
-//-----------------------------------------------------------------------------
-static PyObject *Connection_GetNationalEncoding(
-    udt_Connection *self,               // connection object
-    void *arg)                          // optional argument (ignored)
-{
-    return Connection_GetCharacterSetName(self, OCI_ATTR_ENV_NCHARSET_ID);
-}
-#endif
-
-
 //-----------------------------------------------------------------------------
 // Connection_GetStmtCacheSize()
 //   Return the Oracle statement cache size.
@@ -1120,7 +1128,7 @@ static PyObject *Connection_GetVersion(
     }
 
     // call stored procedure
-    if (Cursor_Call(cursor, NULL, procName, listOfArguments) < 0) {
+    if (Cursor_Call(cursor, NULL, procName, listOfArguments, NULL) < 0) {
         Py_DECREF(procName);
         Py_DECREF(listOfArguments);
         Py_DECREF(cursor);
@@ -1137,6 +1145,31 @@ static PyObject *Connection_GetVersion(
 }
 
 
+//-----------------------------------------------------------------------------
+// Connection_GetEncoding()
+//   Return the encoding associated with the environment of the connection.
+//-----------------------------------------------------------------------------
+static PyObject *Connection_GetEncoding(
+    udt_Connection *self,               // connection object
+    void *arg)                          // optional argument (ignored)
+{
+    return cxString_FromAscii(self->environment->encoding);
+}
+
+
+//-----------------------------------------------------------------------------
+// Connection_GetNationalEncoding()
+//   Return the national encoding associated with the environment of the
+// connection.
+//-----------------------------------------------------------------------------
+static PyObject *Connection_GetNationalEncoding(
+    udt_Connection *self,               // connection object
+    void *arg)                          // optional argument (ignored)
+{
+    return cxString_FromAscii(self->environment->nencoding);
+}
+
+
 //-----------------------------------------------------------------------------
 // Connection_GetMaxBytesPerCharacter()
 //   Return the maximum number of bytes per character.
@@ -1184,6 +1217,14 @@ static PyObject *Connection_Close(
         OCIHandleFree(self->handle, OCI_HTYPE_SVCCTX);
     }
     self->handle = NULL;
+    if (self->serverHandle) {
+        status = OCIServerDetach(self->serverHandle,
+                self->environment->errorHandle, OCI_DEFAULT);
+        if (Environment_CheckForError(self->environment, status,
+                "Connection_Close(): server detach") < 0)
+            return NULL;
+        self->serverHandle = NULL;
+    }
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -1542,7 +1583,8 @@ static PyObject *Connection_ContextManagerExit(
 }
 
 
-#ifdef ORACLE_10GR2
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
+#if !defined(AIX5) || ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
 //-----------------------------------------------------------------------------
 // Connection_Ping()
 //   Makes a round trip call to the server to confirm that the connection and
@@ -1564,6 +1606,7 @@ static PyObject *Connection_Ping(
     Py_INCREF(Py_None);
     return Py_None;
 }
+#endif
 
 
 //-----------------------------------------------------------------------------
@@ -1662,20 +1705,20 @@ static PyObject *Connection_Subscribe(
     PyObject* keywordArgs)              // keyword arguments
 {
     static char *keywordList[] = { "namespace", "protocol", "callback",
-            "timeout", "operations", "rowids", NULL };
-    ub4 namespace, protocol, timeout, rowids, operations;
+            "timeout", "operations", "rowids", "port", "qos", "cqqos", NULL };
+    ub4 namespace, protocol, port, timeout, rowids, operations, qos, cqqos;
     PyObject *rowidsObj, *callback;
     int temp;
 
     // parse arguments
-    timeout = rowids = 0;
+    timeout = rowids = port = qos = cqqos = 0;
     rowidsObj = callback = NULL;
     namespace = OCI_SUBSCR_NAMESPACE_DBCHANGE;
     protocol = OCI_SUBSCR_PROTO_OCI;
     operations = OCI_OPCODE_ALLOPS;
-    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "|iiOiiO", keywordList,
+    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "|iiOiiOiii", keywordList,
             &namespace, &protocol, &callback, &timeout, &operations,
-            &rowidsObj))
+            &rowidsObj, &port, &qos, &cqqos))
         return NULL;
 
     // set the value for rowids
@@ -1687,8 +1730,8 @@ static PyObject *Connection_Subscribe(
             rowids = 1;
     }
 
-    return (PyObject*) Subscription_New(self, namespace, protocol, callback,
-            timeout, operations, rowids);
+    return (PyObject*) Subscription_New(self, namespace, protocol, port,
+            callback, timeout, operations, qos, cqqos, rowids);
 }
 #endif
 
