@@ -135,7 +135,7 @@ static PyGetSetDef g_CursorCalcMembers[] = {
 //-----------------------------------------------------------------------------
 static PyTypeObject g_CursorType = {
     PyVarObject_HEAD_INIT(NULL, 0)
-    "OracleCursor",                     // tp_name
+    "cx_Oracle.Cursor",                 // tp_name
     sizeof(udt_Cursor),                 // tp_basicsize
     0,                                  // tp_itemsize
     (destructor) Cursor_Free,           // tp_dealloc
@@ -206,23 +206,28 @@ static int Cursor_FreeHandle(
     udt_Cursor *self,                   // cursor object
     int raiseException)                 // raise an exception, if necesary?
 {
-    udt_StringBuffer buffer;
+    udt_Buffer buffer;
     sword status;
 
     if (self->handle) {
         if (self->isOwned) {
-            OCIHandleFree(self->handle, OCI_HTYPE_STMT);
+            status = OCIHandleFree(self->handle, OCI_HTYPE_STMT);
+            if (raiseException && Environment_CheckForError(
+                    self->environment, status, "Cursor_FreeHandle()") < 0)
+                return -1;
         } else if (self->connection->handle != 0) {
-            if (!StringBuffer_Fill(&buffer, self->statementTag) < 0)
+            if (!cxBuffer_FromObject(&buffer, self->statementTag,
+                    self->environment->encoding) < 0)
                 return (raiseException) ? -1 : 0;
             status = OCIStmtRelease(self->handle,
                     self->environment->errorHandle, (text*) buffer.ptr,
                     buffer.size, OCI_DEFAULT);
-            StringBuffer_Clear(&buffer);
+            cxBuffer_Clear(&buffer);
             if (raiseException && Environment_CheckForError(
                     self->environment, status, "Cursor_FreeHandle()") < 0)
                 return -1;
         }
+        self->handle = NULL;
     }
     return 0;
 }
@@ -424,7 +429,8 @@ static int Cursor_GetBindNames(
     for (i = 0; i < foundElements; i++) {
         if (!duplicate[i]) {
             temp = cxString_FromEncodedString(bindNames[i],
-                    bindNameLengths[i]);
+                    bindNameLengths[i],
+                    self->connection->environment->encoding);
             if (!temp) {
                 Py_DECREF(*names);
                 PyMem_Free(buffer);
@@ -498,7 +504,9 @@ static int Cursor_SetRowCount(
         self->rowNum = 0;
     } else if (self->statementType == OCI_STMT_INSERT ||
                self->statementType == OCI_STMT_UPDATE ||
-               self->statementType == OCI_STMT_DELETE) {
+               self->statementType == OCI_STMT_DELETE ||
+               self->statementType == OCI_STMT_BEGIN ||
+               self->statementType == OCI_STMT_DECLARE) {
         status = OCIAttrGet(self->handle, OCI_HTYPE_STMT, &rowCount, 0,
                 OCI_ATTR_ROW_COUNT, self->environment->errorHandle);
         if (Environment_CheckForError(self->environment, status,
@@ -520,14 +528,19 @@ static int Cursor_SetRowCount(
 static void Cursor_SetErrorOffset(
     udt_Cursor *self)                   // cursor to get the error offset from
 {
-    PyObject *type, *value, *traceback;
+    PyObject *type, *value, *traceback, *args;
     udt_Error *error;
+    ub2 offset = 0;
 
     PyErr_Fetch(&type, &value, &traceback);
     if (type == g_DatabaseErrorException) {
-        error = (udt_Error*) value;
-        OCIAttrGet(self->handle, OCI_HTYPE_STMT, &error->offset, 0,
+        PyErr_NormalizeException(&type, &value, &traceback);
+        OCIAttrGet(self->handle, OCI_HTYPE_STMT, &offset, 0,
                 OCI_ATTR_PARSE_ERROR_OFFSET, self->environment->errorHandle);
+        args = PyObject_GetAttrString(value, "args");
+        error = (udt_Error*) PyTuple_GET_ITEM(args, 0);
+        error->offset = offset;
+        Py_DECREF(args);
     }
     PyErr_Restore(type, value, traceback);
 }
@@ -621,10 +634,10 @@ static PyObject *Cursor_ItemDescriptionHelper(
     unsigned pos,                       // position in description
     OCIParam *param)                    // parameter to use for description
 {
+    ub2 internalSize, charSize;
     udt_VariableType *varType;
     int displaySize, index;
     PyObject *tuple, *type;
-    ub2 internalSize;
     ub4 nameLength;
     sb2 precision;
     sword status;
@@ -644,6 +657,13 @@ static PyObject *Cursor_ItemDescriptionHelper(
             "Cursor_ItemDescription(): internal size") < 0)
         return NULL;
 
+    // acquire character size of item
+    status = OCIAttrGet(param, OCI_HTYPE_DESCRIBE, (dvoid*) &charSize, 0,
+            OCI_ATTR_CHAR_SIZE, self->environment->errorHandle);
+    if (Environment_CheckForError(self->environment, status,
+            "Cursor_ItemDescription(): character size") < 0)
+        return NULL;
+
     // aquire name of item
     status = OCIAttrGet(param, OCI_HTYPE_DESCRIBE, (dvoid*) &name,
             &nameLength, OCI_ATTR_NAME, self->environment->errorHandle);
@@ -677,18 +697,18 @@ static PyObject *Cursor_ItemDescriptionHelper(
     // set display size based on data type
     type = (PyObject*) varType->pythonType;
     if (type == (PyObject*) &g_StringVarType)
-        displaySize = internalSize;
-#ifndef WITH_UNICODE
+        displaySize = charSize;
+#if PY_MAJOR_VERSION < 3
     else if (type == (PyObject*) &g_UnicodeVarType)
-        displaySize = internalSize / 2;
+        displaySize = charSize;
 #endif
     else if (type == (PyObject*) &g_BinaryVarType)
         displaySize = internalSize;
     else if (type == (PyObject*) &g_FixedCharVarType)
-        displaySize = internalSize;
-#ifndef WITH_UNICODE
+        displaySize = charSize;
+#if PY_MAJOR_VERSION < 3
     else if (type == (PyObject*) &g_FixedUnicodeVarType)
-        displaySize = internalSize / 2;
+        displaySize = charSize;
 #endif
     else if (type == (PyObject*) &g_NumberVarType) {
         if (precision) {
@@ -709,7 +729,8 @@ static PyObject *Cursor_ItemDescriptionHelper(
         return NULL;
 
     // set each of the items in the tuple
-    PyTuple_SET_ITEM(tuple, 0, cxString_FromEncodedString(name, nameLength));
+    PyTuple_SET_ITEM(tuple, 0, cxString_FromEncodedString(name, nameLength,
+            self->connection->environment->encoding));
     Py_INCREF(type);
     PyTuple_SET_ITEM(tuple, 1, type);
     PyTuple_SET_ITEM(tuple, 2, PyInt_FromLong(displaySize));
@@ -825,7 +846,6 @@ static PyObject *Cursor_Close(
     if (Cursor_FreeHandle(self, 1) < 0)
         return NULL;
 
-    self->handle = NULL;
     self->isOpen = 0;
 
     Py_INCREF(Py_None);
@@ -1031,6 +1051,12 @@ static int Cursor_PerformBind(
     Py_ssize_t pos;
     ub2 i;
 
+    // ensure that input sizes are reset
+    // this is done before binding is attempted so that if binding fails and
+    // a new statement is prepared, the bind variables will be reset and
+    // spurious errors will not occur
+    self->setInputSizes = 0;
+
     // set values and perform binds for all bind variables
     if (self->bindVariables) {
         if (PyDict_Check(self->bindVariables)) {
@@ -1051,9 +1077,6 @@ static int Cursor_PerformBind(
         }
     }
 
-    // ensure that input sizes are reset
-    self->setInputSizes = 0;
-
     return 0;
 }
 
@@ -1113,7 +1136,7 @@ static int Cursor_InternalPrepare(
     PyObject *statement,                // statement to prepare
     PyObject *statementTag)             // tag of statement to prepare
 {
-    udt_StringBuffer statementBuffer, tagBuffer;
+    udt_Buffer statementBuffer, tagBuffer;
     sword status;
 
     // make sure we don't get a situation where nothing is to be executed
@@ -1147,10 +1170,12 @@ static int Cursor_InternalPrepare(
 
     // prepare statement
     self->isOwned = 0;
-    if (StringBuffer_Fill(&statementBuffer, statement) < 0)
+    if (cxBuffer_FromObject(&statementBuffer, statement,
+            self->environment->encoding) < 0)
         return -1;
-    if (StringBuffer_Fill(&tagBuffer, statementTag) < 0) {
-        StringBuffer_Clear(&statementBuffer);
+    if (cxBuffer_FromObject(&tagBuffer, statementTag,
+            self->environment->encoding) < 0) {
+        cxBuffer_Clear(&statementBuffer);
         return -1;
     }
     Py_BEGIN_ALLOW_THREADS
@@ -1159,8 +1184,8 @@ static int Cursor_InternalPrepare(
             statementBuffer.size, (text*) tagBuffer.ptr, tagBuffer.size,
             OCI_NTV_SYNTAX, OCI_DEFAULT);
     Py_END_ALLOW_THREADS
-    StringBuffer_Clear(&statementBuffer);
-    StringBuffer_Clear(&tagBuffer);
+    cxBuffer_Clear(&statementBuffer);
+    cxBuffer_Clear(&tagBuffer);
     if (Environment_CheckForError(self->environment, status,
             "Cursor_InternalPrepare(): prepare") < 0) {
         // this is needed to avoid "invalid handle" errors since Oracle doesn't
@@ -1242,8 +1267,7 @@ static PyObject *Cursor_Prepare(
 
     // statement text and optional tag is expected
     statementTag = NULL;
-    if (!PyArg_ParseTuple(args, "O!|O!", cxString_Type, &statement,
-            cxString_Type, &statementTag))
+    if (!PyArg_ParseTuple(args, "O|O", &statement, &statementTag))
         return NULL;
 
     // make sure the cursor is open
@@ -1259,6 +1283,164 @@ static PyObject *Cursor_Prepare(
 }
 
 
+//-----------------------------------------------------------------------------
+// Cursor_CallCalculateSize()
+//   Calculate the size of the statement that is to be executed.
+//-----------------------------------------------------------------------------
+static int Cursor_CallCalculateSize(
+    PyObject *name,                     // name of procedure/function to call
+    udt_Variable *returnValue,          // return value variable (optional)
+    PyObject *listOfArguments,          // list of positional arguments
+    PyObject *keywordArguments,         // dictionary of keyword arguments
+    int *size)                          // statement size (OUT)
+{
+    int numPositionalArgs, numKeywordArgs;
+
+    // set base size without any arguments
+    *size = 17;
+
+    // add any additional space required to handle the return value
+    if (returnValue)
+        *size += 6;
+
+    // assume up to 9 characters for each positional argument
+    // this allows up to four digits for the placeholder if the bind variale
+    // is a boolean value
+    if (listOfArguments) {
+        numPositionalArgs = PySequence_Size(listOfArguments);
+        if (numPositionalArgs < 0)
+            return -1;
+        *size += numPositionalArgs * 9;
+    }
+
+    // assume up to 15 characters for each keyword argument
+    // this allows up to four digits for the placeholder if the bind variable
+    // is a boolean value
+    if (keywordArguments) {
+        numKeywordArgs = PyDict_Size(keywordArguments);
+        if (numKeywordArgs < 0)
+            return -1;
+        *size += numKeywordArgs * 15;
+    }
+
+    return 0;
+}
+
+
+//-----------------------------------------------------------------------------
+// Cursor_CallBuildStatement()
+//   Determine the statement and the bind variables to bind to the statement
+// that is created for calling a stored procedure or function.
+//-----------------------------------------------------------------------------
+static int Cursor_CallBuildStatement(
+    PyObject *name,                     // name of procedure/function to call
+    udt_Variable *returnValue,          // return value variable (optional)
+    PyObject *listOfArguments,          // arguments
+    PyObject *keywordArguments,         // keyword arguments
+    char *statement,                    // allocated statement text
+    PyObject **statementObj,            // statement object (OUT)
+    PyObject **bindVariables)           // variables to bind (OUT)
+{
+    PyObject *key, *value, *format, *formatArgs, *positionalArgs, *temp;
+    int i, argNum, numPositionalArgs;
+    Py_ssize_t pos;
+    char *ptr;
+
+    // initialize the bind variables to the list of positional arguments
+    if (listOfArguments)
+        *bindVariables = PySequence_List(listOfArguments);
+    else *bindVariables = PyList_New(0);
+    if (!*bindVariables)
+        return -1;
+
+    // insert the return variable, if applicable
+    if (returnValue) {
+        if (PyList_Insert(*bindVariables, 0, (PyObject*) returnValue) < 0)
+            return -1;
+    }
+
+    // initialize format arguments
+    formatArgs = PyList_New(0);
+    if (!formatArgs)
+        return -1;
+    if (PyList_Append(formatArgs, name) < 0) {
+        Py_DECREF(formatArgs);
+        return -1;
+    }
+
+    // begin building the statement
+    argNum = 1;
+    strcpy(statement, "begin ");
+    if (returnValue) {
+        strcat(statement, ":1 := ");
+        argNum++;
+    }
+    strcat(statement, "%s");
+    ptr = statement + strlen(statement);
+    *ptr++ = '(';
+
+    // include any positional arguments first
+    if (listOfArguments) {
+        positionalArgs = PySequence_Fast(listOfArguments,
+                "expecting sequence of arguments");
+        if (!positionalArgs) {
+            Py_DECREF(formatArgs);
+            return -1;
+        }
+        numPositionalArgs = PySequence_Size(listOfArguments);
+        for (i = 0; i < numPositionalArgs; i++) {
+            if (i > 0)
+                *ptr++ = ',';
+            ptr += sprintf(ptr, ":%d", argNum++);
+            if (PyBool_Check(PySequence_Fast_GET_ITEM(positionalArgs, i)))
+                ptr += sprintf(ptr, " = 1");
+        }
+        Py_DECREF(positionalArgs);
+    }
+
+    // next append any keyword arguments
+    if (keywordArguments) {
+        pos = 0;
+        while (PyDict_Next(keywordArguments, &pos, &key, &value)) {
+            if (PyList_Append(*bindVariables, value) < 0) {
+                Py_DECREF(formatArgs);
+                return -1;
+            }
+            if (PyList_Append(formatArgs, key) < 0) {
+                Py_DECREF(formatArgs);
+                return -1;
+            }
+            if ((argNum > 1 && !returnValue) || (argNum > 2 && returnValue))
+                *ptr++ = ',';
+            ptr += sprintf(ptr, "%%s => :%d", argNum++);
+            if (PyBool_Check(value))
+                ptr += sprintf(ptr, " = 1");
+        }
+    }
+
+    // create statement object
+    strcpy(ptr, "); end;");
+    format = cxString_FromAscii(statement);
+    if (!format) {
+        Py_DECREF(formatArgs);
+        return -1;
+    }
+    temp = PyList_AsTuple(formatArgs);
+    Py_DECREF(formatArgs);
+    if (!temp) {
+        Py_DECREF(format);
+        return -1;
+    }
+    *statementObj = cxString_Format(format, temp);
+    Py_DECREF(format);
+    Py_DECREF(temp);
+    if (!*statementObj)
+        return -1;
+
+    return 0;
+}
+
+
 //-----------------------------------------------------------------------------
 // Cursor_Call()
 //   Call a stored procedure or function.
@@ -1267,105 +1449,54 @@ static int Cursor_Call(
     udt_Cursor *self,                   // cursor to call procedure/function
     udt_Variable *returnValue,          // return value variable (optional)
     PyObject *name,                     // name of procedure/function to call
-    PyObject *listOfArguments)          // arguments to procedure/function
+    PyObject *listOfArguments,          // arguments
+    PyObject *keywordArguments)         // keyword arguments
 {
-    PyObject *bindVariables, *results, *arguments, *statementObj;
-    int numArguments, statementSize, i, offset;
-    PyObject *format, *formatArgs;
-    char *statement, *ptr;
+    PyObject *bindVariables, *statementObj, *results;
+    int statementSize;
+    char *statement;
 
-    // determine the number of arguments passed
+    // verify that the arguments are passed correctly
     if (listOfArguments) {
         if (!PySequence_Check(listOfArguments)) {
             PyErr_SetString(PyExc_TypeError, "arguments must be a sequence");
             return -1;
         }
-        numArguments = PySequence_Size(listOfArguments);
-        if (numArguments < 0)
-            return -1;
-    } else {
-        numArguments = 0;
-        listOfArguments = PyList_New(0);
-        if (!listOfArguments)
+    }
+    if (keywordArguments) {
+        if (!PyDict_Check(keywordArguments)) {
+            PyErr_SetString(PyExc_TypeError,
+                    "keyword arguments must be a dictionary");
             return -1;
+        }
     }
 
     // make sure the cursor is open
     if (Cursor_IsOpen(self) < 0)
         return -1;
 
-    // determine the statement size add the return value, if applicable
-    offset = 0;
-    statementSize = numArguments * 9 + 17;
-    if (returnValue) {
-        offset = 1;
-        statementSize += 10;
-        bindVariables = PySequence_List(listOfArguments);
-        if (!bindVariables)
-            return -1;
-        if (PyList_Insert(bindVariables, 0, (PyObject*) returnValue) < 0) {
-            Py_DECREF(bindVariables);
-            return -1;
-        }
-    } else {
-        Py_INCREF(listOfArguments);
-        bindVariables = listOfArguments;
-    }
+    // determine the statement size
+    if (Cursor_CallCalculateSize(name, returnValue, listOfArguments,
+            keywordArguments, &statementSize) < 0)
+        return -1;
 
     // allocate a string for the statement
     statement = (char*) PyMem_Malloc(statementSize);
     if (!statement) {
-        Py_DECREF(bindVariables);
         PyErr_NoMemory();
         return -1;
     }
 
-    // build up the statement
-    arguments = PySequence_Fast(listOfArguments,
-            "expecting sequence of arguments");
-    if (!arguments) {
-        Py_DECREF(bindVariables);
+    // determine the statement to execute and the argument to pass
+    bindVariables = statementObj = NULL;
+    if (Cursor_CallBuildStatement(name, returnValue, listOfArguments,
+            keywordArguments, statement, &statementObj, &bindVariables) < 0) {
         PyMem_Free(statement);
+        Py_XDECREF(statementObj);
+        Py_XDECREF(bindVariables);
         return -1;
     }
-    strcpy(statement, "begin ");
-    if (returnValue)
-        strcat(statement, ":1 := ");
-    strcat(statement, "%s");
-    ptr = statement + strlen(statement);
-    *ptr++ = '(';
-    for (i = 0; i < numArguments; i++) {
-        if (i > 0)
-            *ptr++ = ',';
-        ptr += sprintf(ptr, ":%d", i + offset + 1);
-        if (PyBool_Check(PySequence_Fast_GET_ITEM(arguments, i)))
-            ptr += sprintf(ptr, " = 1");
-    }
-    strcpy(ptr, "); end;");
-    Py_DECREF(arguments);
-
-    // create the statement object
-    format = cxString_FromAscii(statement);
     PyMem_Free(statement);
-    if (!format) {
-        Py_DECREF(bindVariables);
-        return -1;
-    }
-    formatArgs = PyTuple_New(1);
-    if (!formatArgs) {
-        Py_DECREF(bindVariables);
-        Py_DECREF(format);
-        return -1;
-    }
-    Py_INCREF(name);
-    PyTuple_SET_ITEM(formatArgs, 0, name);
-    statementObj = cxString_Format(format, formatArgs);
-    Py_DECREF(format);
-    Py_DECREF(formatArgs);
-    if (!statementObj) {
-        Py_DECREF(bindVariables);
-        return -1;
-    }
 
     // execute the statement on the cursor
     results = PyObject_CallMethod( (PyObject*) self, "execute", "OO",
@@ -1389,15 +1520,15 @@ static PyObject *Cursor_CallFunc(
     PyObject *args,                     // arguments
     PyObject *keywordArgs)              // keyword arguments
 {
-    static char *keywordList[] = { "name", "returnType", "parameters", NULL };
-    PyObject *listOfArguments, *returnType, *results, *name;
+    static char *keywordList[] = { "name", "returnType", "parameters",
+            "keywordParameters", NULL };
+    PyObject *listOfArguments, *keywordArguments, *returnType, *results, *name;
     udt_Variable *var;
 
-    // expect stored function name, return type and optionally a list of
-    // arguments
-    listOfArguments = NULL;
-    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "O!O|O", keywordList,
-            cxString_Type, &name, &returnType, &listOfArguments))
+    // parse arguments
+    listOfArguments = keywordArguments = NULL;
+    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "OO|OO", keywordList,
+            &name, &returnType, &listOfArguments, &keywordArguments))
         return NULL;
 
     // create the return variable
@@ -1406,7 +1537,7 @@ static PyObject *Cursor_CallFunc(
         return NULL;
 
     // call the function
-    if (Cursor_Call(self, var, name, listOfArguments) < 0)
+    if (Cursor_Call(self, var, name, listOfArguments, keywordArguments) < 0)
         return NULL;
 
     // determine the results
@@ -1425,18 +1556,19 @@ static PyObject *Cursor_CallProc(
     PyObject *args,                     // arguments
     PyObject *keywordArgs)              // keyword arguments
 {
-    static char *keywordList[] = { "name", "parameters", NULL };
-    PyObject *listOfArguments, *results, *var, *temp, *name;
+    static char *keywordList[] = { "name", "parameters", "keywordParameters",
+            NULL };
+    PyObject *listOfArguments, *keywordArguments, *results, *var, *temp, *name;
     int numArgs, i;
 
-    // expect stored procedure name and optionally a list of arguments
-    listOfArguments = NULL;
-    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "O!|O", keywordList,
-            cxString_Type, &name, &listOfArguments))
+    // parse arguments
+    listOfArguments = keywordArguments = NULL;
+    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "O|OO", keywordList,
+            &name, &listOfArguments, &keywordArguments))
         return NULL;
 
     // call the stored procedure
-    if (Cursor_Call(self, NULL, name, listOfArguments) < 0)
+    if (Cursor_Call(self, NULL, name, listOfArguments, keywordArguments) < 0)
         return NULL;
 
     // create the return value
@@ -1473,10 +1605,6 @@ static PyObject *Cursor_Execute(
     executeArgs = NULL;
     if (!PyArg_ParseTuple(args, "O|O", &statement, &executeArgs))
         return NULL;
-    if (statement != Py_None && !cxString_Check(statement)) {
-        PyErr_SetString(PyExc_TypeError, "expecting None or a string");
-        return NULL;
-    }
     if (executeArgs && keywordArgs) {
         if (PyDict_Size(keywordArgs) == 0)
             keywordArgs = NULL;
@@ -1552,10 +1680,6 @@ static PyObject *Cursor_ExecuteMany(
     if (!PyArg_ParseTuple(args, "OO!", &statement, &PyList_Type,
             &listOfArguments))
         return NULL;
-    if (statement != Py_None && !cxString_Check(statement)) {
-        PyErr_SetString(PyExc_TypeError, "expecting None or string");
-        return NULL;
-    }
 
     // make sure the cursor is open
     if (Cursor_IsOpen(self) < 0)
@@ -1687,6 +1811,14 @@ static int Cursor_InternalFetch(
         PyErr_SetString(g_InterfaceErrorException, "query not executed");
         return -1;
     }
+    for (i = 0; i < PyList_GET_SIZE(self->fetchVariables); i++) {
+        var = (udt_Variable*) PyList_GET_ITEM(self->fetchVariables, i);
+        var->internalFetchNum++;
+        if (var->type->preFetchProc) {
+            if ((*var->type->preFetchProc)(var) < 0)
+                return -1;
+        }
+    }
     Py_BEGIN_ALLOW_THREADS
     status = OCIStmtFetch(self->handle, self->environment->errorHandle,
             numRows, OCI_FETCH_NEXT, OCI_DEFAULT);
@@ -1696,10 +1828,6 @@ static int Cursor_InternalFetch(
                 "Cursor_InternalFetch(): fetch") < 0)
             return -1;
     }
-    for (i = 0; i < PyList_GET_SIZE(self->fetchVariables); i++) {
-        var = (udt_Variable*) PyList_GET_ITEM(self->fetchVariables, i);
-        var->internalFetchNum++;
-    }
     status = OCIAttrGet(self->handle, OCI_HTYPE_STMT, &rowCount, 0,
             OCI_ATTR_ROW_COUNT, self->environment->errorHandle);
     if (Environment_CheckForError(self->environment, status,
@@ -1977,9 +2105,10 @@ static PyObject *Cursor_Var(
     PyObject *keywordArgs)              // keyword arguments
 {
     static char *keywordList[] = { "type", "size", "arraysize",
-            "inconverter", "outconverter", NULL };
-    PyObject *inConverter, *outConverter;
+            "inconverter", "outconverter", "typename", NULL };
+    PyObject *inConverter, *outConverter, *typeNameObj;
     udt_VariableType *varType;
+    udt_ObjectVar *objectVar;
     int size, arraySize;
     udt_Variable *var;
     PyObject *type;
@@ -1987,9 +2116,10 @@ static PyObject *Cursor_Var(
     // parse arguments
     size = 0;
     arraySize = self->bindArraySize;
-    inConverter = outConverter = NULL;
-    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "O|iiOO", keywordList,
-            &type, &size, &arraySize, &inConverter, &outConverter))
+    inConverter = outConverter = typeNameObj = NULL;
+    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "O|iiOOO", keywordList,
+            &type, &size, &arraySize, &inConverter, &outConverter,
+            &typeNameObj))
         return NULL;
 
     // determine the type of variable
@@ -1998,6 +2128,11 @@ static PyObject *Cursor_Var(
         return NULL;
     if (varType->isVariableLength && size == 0)
         size = varType->size;
+    if (type == (PyObject*) &g_ObjectVarType && !typeNameObj) {
+        PyErr_SetString(PyExc_TypeError,
+                "expecting type name for object variables");
+        return NULL;
+    }
 
     // create the variable
     var = Variable_New(self, arraySize, varType, size);
@@ -2008,6 +2143,17 @@ static PyObject *Cursor_Var(
     Py_XINCREF(outConverter);
     var->outConverter = outConverter;
 
+    // define the object type if needed
+    if (type == (PyObject*) &g_ObjectVarType) {
+        objectVar = (udt_ObjectVar*) var;
+        objectVar->objectType = ObjectType_NewByName(self->connection,
+                typeNameObj);
+        if (!objectVar->objectType) {
+            Py_DECREF(var);
+            return NULL;
+        }
+    }
+
     return (PyObject*) var;
 }
 
