@@ -95,6 +95,7 @@ static udt_VariableType vt_Float = {
     (FinalizeProc) NULL,
     (PreDefineProc) NumberVar_PreDefine,
     (PostDefineProc) NULL,
+    (PreFetchProc) NULL,
     (IsNullProc) NULL,
     (SetValueProc) NumberVar_SetValue,
     (GetValueProc) NumberVar_GetValue,
@@ -116,6 +117,7 @@ static udt_VariableType vt_NativeFloat = {
     (FinalizeProc) NULL,
     (PreDefineProc) NULL,
     (PostDefineProc) NULL,
+    (PreFetchProc) NULL,
     (IsNullProc) NULL,
     (SetValueProc) NativeFloatVar_SetValue,
     (GetValueProc) NativeFloatVar_GetValue,
@@ -132,12 +134,12 @@ static udt_VariableType vt_NativeFloat = {
 #endif
 
 
-#if PY_MAJOR_VERSION < 3
 static udt_VariableType vt_Integer = {
     (InitializeProc) NULL,
     (FinalizeProc) NULL,
     (PreDefineProc) NumberVar_PreDefine,
     (PostDefineProc) NULL,
+    (PreFetchProc) NULL,
     (IsNullProc) NULL,
     (SetValueProc) NumberVar_SetValue,
     (GetValueProc) NumberVar_GetValue,
@@ -151,7 +153,6 @@ static udt_VariableType vt_Integer = {
     1,                                  // can be copied
     1                                   // can be in array
 };
-#endif
 
 
 static udt_VariableType vt_LongInteger = {
@@ -159,6 +160,7 @@ static udt_VariableType vt_LongInteger = {
     (FinalizeProc) NULL,
     (PreDefineProc) NumberVar_PreDefine,
     (PostDefineProc) NULL,
+    (PreFetchProc) NULL,
     (IsNullProc) NULL,
     (SetValueProc) NumberVar_SetValue,
     (GetValueProc) NumberVar_GetValue,
@@ -179,6 +181,7 @@ static udt_VariableType vt_NumberAsString = {
     (FinalizeProc) NULL,
     (PreDefineProc) NumberVar_PreDefine,
     (PostDefineProc) NULL,
+    (PreFetchProc) NULL,
     (IsNullProc) NULL,
     (SetValueProc) NumberVar_SetValue,
     (GetValueProc) NumberVar_GetValue,
@@ -199,6 +202,7 @@ static udt_VariableType vt_Boolean = {
     (FinalizeProc) NULL,
     (PreDefineProc) NumberVar_PreDefine,
     (PostDefineProc) NULL,
+    (PreFetchProc) NULL,
     (IsNullProc) NULL,
     (SetValueProc) NumberVar_SetValue,
     (GetValueProc) NumberVar_GetValue,
@@ -223,6 +227,7 @@ static int NumberVar_PreDefine(
     udt_NumberVar *var,                 // variable to initialize
     OCIParam *param)                    // parameter handle
 {
+    static int maxLongSafeDigits = sizeof(long) >= 8 ? 18 : 9;
     sb2 precision;
     sword status;
     sb1 scale;
@@ -244,10 +249,8 @@ static int NumberVar_PreDefine(
             return -1;
         if (scale == 0 || (scale == -127 && precision == 0)) {
             var->type = &vt_LongInteger;
-#if PY_MAJOR_VERSION < 3
-        if (precision > 0 && precision < 10)
-            var->type = &vt_Integer;
-#endif
+            if (precision > 0 && precision <= maxLongSafeDigits)
+                var->type = &vt_Integer;
         }
     }
 
@@ -326,20 +329,22 @@ static int NumberVar_SetValueFromLong(
     unsigned pos,                       // array position to set
     PyObject *value)                    // value to set
 {
-    udt_StringBuffer textBuffer;
+    udt_Buffer textBuffer;
     PyObject *textValue;
     sword status;
 
-    textValue = cxString_FromObject(value);
+    textValue = PyObject_Str(value);
     if (!textValue)
         return -1;
-    if (StringBuffer_Fill(&textBuffer, textValue) < 0)
+    if (cxBuffer_FromObject(&textBuffer, textValue,
+            var->environment->encoding) < 0)
         return -1;
     status = OCINumberFromText(var->environment->errorHandle,
             (text*) textBuffer.ptr, textBuffer.size,
-            (text*) g_NumberToStringFormatBuffer.ptr,
-            g_NumberToStringFormatBuffer.size, NULL, 0, &var->data[pos]);
-    StringBuffer_Clear(&textBuffer);
+            (text*) var->environment->numberFromStringFormatBuffer.ptr,
+            var->environment->numberFromStringFormatBuffer.size, NULL, 0,
+            &var->data[pos]);
+    cxBuffer_Clear(&textBuffer);
     Py_DECREF(textValue);
     return Environment_CheckForError(var->environment, status,
             "NumberVar_SetValueFromLong()");
@@ -444,7 +449,7 @@ static int NumberVar_SetValueFromDecimal(
     PyObject *value)                    // value to set
 {
     PyObject *textValue, *format, *tupleValue;
-    udt_StringBuffer textBuffer, formatBuffer;
+    udt_Buffer textBuffer, formatBuffer;
     sword status;
 
     tupleValue = PyObject_CallMethod(value, "as_tuple", NULL);
@@ -456,17 +461,22 @@ static int NumberVar_SetValueFromDecimal(
         return -1;
     }
     Py_DECREF(tupleValue);
-    if (StringBuffer_Fill(&textBuffer, textValue) < 0)
+    if (cxBuffer_FromObject(&textBuffer, textValue,
+            var->environment->encoding) < 0)
         return -1;
-    if (StringBuffer_Fill(&formatBuffer, format) < 0) {
-        StringBuffer_Clear(&textBuffer);
+    if (cxBuffer_FromObject(&formatBuffer, format,
+            var->environment->encoding) < 0) {
+        cxBuffer_Clear(&textBuffer);
         return -1;
     }
     status = OCINumberFromText(var->environment->errorHandle,
             (text*) textBuffer.ptr, textBuffer.size, (text*) formatBuffer.ptr,
-            formatBuffer.size, NULL, 0, &var->data[pos]);
-    StringBuffer_Clear(&textBuffer);
-    StringBuffer_Clear(&formatBuffer);
+            formatBuffer.size,
+            var->environment->nlsNumericCharactersBuffer.ptr,
+            var->environment->nlsNumericCharactersBuffer.size,
+            &var->data[pos]);
+    cxBuffer_Clear(&textBuffer);
+    cxBuffer_Clear(&formatBuffer);
     Py_DECREF(textValue);
     Py_DECREF(format);
     return Environment_CheckForError(var->environment, status,
@@ -514,33 +524,34 @@ static PyObject *NumberVar_GetValue(
     ub4 stringLength;
     sword status;
 
-#if PY_MAJOR_VERSION < 3
-    if (var->type == &vt_Integer || var->type == &vt_Boolean) {
-#else
-    if (var->type == &vt_Boolean) {
-#endif
+    if (var->type == &vt_Boolean || var->type == &vt_Integer) {
         status = OCINumberToInt(var->environment->errorHandle, &var->data[pos],
                 sizeof(long), OCI_NUMBER_SIGNED, (dvoid*) &integerValue);
         if (Environment_CheckForError(var->environment, status,
                 "NumberVar_GetValue(): as integer") < 0)
             return NULL;
-#if PY_MAJOR_VERSION < 3
-        if (var->type == &vt_Integer)
-            return PyInt_FromLong(integerValue);
+
+        if (var->type == &vt_Boolean)
+            return PyBool_FromLong(integerValue);
+#if PY_MAJOR_VERSION >= 3
+        return PyLong_FromLong(integerValue);
+#else
+        return PyInt_FromLong(integerValue);
 #endif
-        return PyBool_FromLong(integerValue);
     }
 
     if (var->type == &vt_NumberAsString || var->type == &vt_LongInteger) {
         stringLength = sizeof(stringValue);
         status = OCINumberToText(var->environment->errorHandle,
-                &var->data[pos], (text*) g_ShortNumberToStringFormatBuffer.ptr,
-                g_ShortNumberToStringFormatBuffer.size, NULL, 0, &stringLength,
-                (unsigned char*) stringValue);
+                &var->data[pos],
+                (text*) var->environment->numberToStringFormatBuffer.ptr,
+                var->environment->numberToStringFormatBuffer.size, NULL, 0,
+                &stringLength, (unsigned char*) stringValue);
         if (Environment_CheckForError(var->environment, status,
                 "NumberVar_GetValue(): as string") < 0)
             return NULL;
-        stringObj = cxString_FromEncodedString(stringValue, stringLength);
+        stringObj = cxString_FromEncodedString(stringValue, stringLength,
+                var->environment->encoding);
         if (!stringObj)
             return NULL;
         if (var->type == &vt_NumberAsString)
