@@ -13,16 +13,22 @@ typedef struct {
     PyObject *callback;
     ub4 namespace;
     ub4 protocol;
+    ub4 port;
     ub4 timeout;
     ub4 operations;
+    ub4 qos;
+    ub4 cqqos;
     ub4 rowids;
+    ub4 id;
 } udt_Subscription;
 
 typedef struct {
     PyObject_HEAD
+    udt_Subscription *subscription;
     ub4 type;
     PyObject *dbname;
     PyObject *tables;
+    PyObject *queries;
 } udt_Message;
 
 typedef struct {
@@ -38,6 +44,15 @@ typedef struct {
     ub4 operation;
 } udt_MessageRow;
 
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+typedef struct {
+    PyObject_HEAD
+    ub8 id;
+    ub4 operation;
+    PyObject *tables;
+} udt_MessageQuery;
+#endif
+
 
 //-----------------------------------------------------------------------------
 // Declaration of subscription functions
@@ -48,6 +63,9 @@ static PyObject *Subscription_RegisterQuery(udt_Subscription*, PyObject*);
 static void Message_Free(udt_Message*);
 static void MessageTable_Free(udt_MessageTable*);
 static void MessageRow_Free(udt_MessageRow*);
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+static void MessageQuery_Free(udt_MessageQuery*);
+#endif
 
 //-----------------------------------------------------------------------------
 // declaration of members for Python types
@@ -58,16 +76,23 @@ static PyMemberDef g_SubscriptionTypeMembers[] = {
             READONLY },
     { "namespace", T_INT, offsetof(udt_Subscription, namespace), READONLY },
     { "protocol", T_INT, offsetof(udt_Subscription, protocol), READONLY },
+    { "port", T_INT, offsetof(udt_Subscription, port), READONLY },
     { "timeout", T_INT, offsetof(udt_Subscription, timeout), READONLY },
     { "operations", T_INT, offsetof(udt_Subscription, operations), READONLY },
+    { "qos", T_INT, offsetof(udt_Subscription, qos), READONLY },
+    { "cqqos", T_INT, offsetof(udt_Subscription, cqqos), READONLY },
     { "rowids", T_BOOL, offsetof(udt_Subscription, rowids), READONLY },
+    { "id", T_INT, offsetof(udt_Subscription, id), READONLY },
     { NULL }
 };
 
 static PyMemberDef g_MessageTypeMembers[] = {
+    { "subscription", T_OBJECT, offsetof(udt_Message, subscription),
+            READONLY },
     { "type", T_INT, offsetof(udt_Message, type), READONLY },
     { "dbname", T_OBJECT, offsetof(udt_Message, dbname), READONLY },
     { "tables", T_OBJECT, offsetof(udt_Message, tables), READONLY },
+    { "queries", T_OBJECT, offsetof(udt_Message, queries), READONLY },
     { NULL }
 };
 
@@ -84,6 +109,15 @@ static PyMemberDef g_MessageRowTypeMembers[] = {
     { NULL }
 };
 
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+static PyMemberDef g_MessageQueryTypeMembers[] = {
+    { "id", T_INT, offsetof(udt_MessageQuery, id), READONLY },
+    { "operation", T_INT, offsetof(udt_MessageQuery, operation), READONLY },
+    { "tables", T_OBJECT, offsetof(udt_MessageQuery, tables), READONLY },
+    { NULL }
+};
+#endif
+
 
 //-----------------------------------------------------------------------------
 // declaration of methods for Python types
@@ -276,6 +310,53 @@ static PyTypeObject g_MessageRowType = {
 };
 
 
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+static PyTypeObject g_MessageQueryType = {
+    PyVarObject_HEAD_INIT(NULL, 0)
+    "cx_Oracle.MessageQuery",           // tp_name
+    sizeof(udt_MessageQuery),           // tp_basicsize
+    0,                                  // tp_itemsize
+    (destructor) MessageQuery_Free,     // tp_dealloc
+    0,                                  // tp_print
+    0,                                  // tp_getattr
+    0,                                  // tp_setattr
+    0,                                  // tp_compare
+    0,                                  // tp_repr
+    0,                                  // tp_as_number
+    0,                                  // tp_as_sequence
+    0,                                  // tp_as_mapping
+    0,                                  // tp_hash
+    0,                                  // tp_call
+    0,                                  // tp_str
+    0,                                  // tp_getattro
+    0,                                  // tp_setattro
+    0,                                  // tp_as_buffer
+    Py_TPFLAGS_DEFAULT,                 // tp_flags
+    0,                                  // tp_doc
+    0,                                  // tp_traverse
+    0,                                  // tp_clear
+    0,                                  // tp_richcompare
+    0,                                  // tp_weaklistoffset
+    0,                                  // tp_iter
+    0,                                  // tp_iternext
+    0,                                  // tp_methods
+    g_MessageQueryTypeMembers,          // tp_members
+    0,                                  // tp_getset
+    0,                                  // tp_base
+    0,                                  // tp_dict
+    0,                                  // tp_descr_get
+    0,                                  // tp_descr_set
+    0,                                  // tp_dictoffset
+    0,                                  // tp_init
+    0,                                  // tp_alloc
+    0,                                  // tp_new
+    0,                                  // tp_free
+    0,                                  // tp_is_gc
+    0                                   // tp_bases
+};
+#endif
+
+
 //-----------------------------------------------------------------------------
 // MessageRow_Initialize()
 //   Initialize a new message row with the information from the descriptor.
@@ -302,7 +383,8 @@ static int MessageRow_Initialize(
     if (Environment_CheckForError(env, status,
             "MessageRow_Initialize(): get rowid") < 0)
         return -1;
-    self->rowid = cxString_FromEncodedString(rowid, rowidLength);
+    self->rowid = cxString_FromEncodedString(rowid, rowidLength,
+            env->encoding);
     if (!self->rowid)
         return -1;
 
@@ -320,11 +402,12 @@ static int MessageTable_Initialize(
     dvoid *descriptor)                  // descriptor to get information from
 {
     dvoid **rowDescriptor, *indicator;
-    ub4 nameLength, i, numRows;
+    ub4 nameLength, i;
     udt_MessageRow *row;
     boolean exists;
     OCIColl *rows;
     sword status;
+    sb4 numRows;
     char *name;
 
     // determine operation
@@ -340,7 +423,7 @@ static int MessageTable_Initialize(
     if (Environment_CheckForError(env, status,
             "MessageTable_Initialize(): get table name") < 0)
         return -1;
-    self->name = cxString_FromEncodedString(name, nameLength);
+    self->name = cxString_FromEncodedString(name, nameLength, env->encoding);
     if (!self->name)
         return -1;
 
@@ -384,6 +467,82 @@ static int MessageTable_Initialize(
 }
 
 
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+//-----------------------------------------------------------------------------
+// MessageQuery_Initialize()
+//   Initialize a new message query with the information from the descriptor.
+//-----------------------------------------------------------------------------
+static int MessageQuery_Initialize(
+    udt_MessageQuery *self,             // object to initialize
+    udt_Environment *env,               // environment to use
+    dvoid *descriptor)                  // descriptor to get information from
+{
+    dvoid **tableDescriptor, *indicator;
+    udt_MessageTable *table;
+    ub4 i;
+    OCIColl *tables;
+    boolean exists;
+    sb4 numTables;
+    sword status;
+
+    // determine query id
+    status = OCIAttrGet(descriptor, OCI_DTYPE_CQDES, &self->id,
+            NULL, OCI_ATTR_CQDES_QUERYID, env->errorHandle);
+    if (Environment_CheckForError(env, status,
+            "MessageQuery_Initialize(): get query id") < 0)
+        return -1;
+
+    // determine operation
+    status = OCIAttrGet(descriptor, OCI_DTYPE_CQDES, &self->operation,
+            NULL, OCI_ATTR_CQDES_OPERATION, env->errorHandle);
+    if (Environment_CheckForError(env, status,
+            "MessageQuery_Initialize(): get operation") < 0)
+        return -1;
+
+    // determine table collection
+    status = OCIAttrGet(descriptor, OCI_DTYPE_CQDES, &tables, NULL,
+            OCI_ATTR_CQDES_TABLE_CHANGES, env->errorHandle);
+    if (Environment_CheckForError(env, status,
+            "MessageQuery_Initialize(): get tables collection") < 0)
+        return -1;
+
+    // determine number of tables
+    if (!tables)
+        numTables = 0;
+    else {
+        status = OCICollSize(env->handle, env->errorHandle, tables,
+                &numTables);
+        if (Environment_CheckForError(env, status,
+                "MessageQuery_Initialize(): get size of collection") < 0)
+            return -1;
+    }
+
+    // create list to hold results
+    self->tables = PyList_New(numTables);
+    if (!self->tables)
+        return -1;
+
+    // populate each entry with a message table instance
+    for (i = 0; i < numTables; i++) {
+        status = OCICollGetElem(env->handle, env->errorHandle, tables, i,
+                &exists, (dvoid*) &tableDescriptor, &indicator);
+        if (Environment_CheckForError(env, status,
+                "MessageQuery_Initialize(): get element from collection") < 0)
+            return -1;
+        table = (udt_MessageTable*)
+                g_MessageTableType.tp_alloc(&g_MessageTableType, 0);
+        if (!table)
+            return -1;
+        PyList_SET_ITEM(self->tables, i, (PyObject*) table);
+        if (MessageTable_Initialize(table, env, *tableDescriptor) < 0)
+            return -1;
+    }
+
+    return 0;
+}
+#endif
+
+
 //-----------------------------------------------------------------------------
 // Message_Initialize()
 //   Initialize a new message with the information from the descriptor.
@@ -391,16 +550,28 @@ static int MessageTable_Initialize(
 static int Message_Initialize(
     udt_Message *self,                  // object to initialize
     udt_Environment *env,               // environment to use
+    udt_Subscription *subscription,     // associated subscription for message
     dvoid *descriptor)                  // descriptor to get information from
 {
     dvoid **tableDescriptor, *indicator;
-    ub4 dbnameLength, numTables, i;
     udt_MessageTable *table;
+    ub4 dbnameLength, i;
     OCIColl *tables;
     boolean exists;
+    sb4 numTables;
     char *dbname;
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+    dvoid **queryDescriptor;
+    udt_MessageQuery *query;
+    OCIColl *queries;
+    sb4 numQueries;
+#endif
     sword status;
 
+    // assign reference to associated subscription
+    Py_INCREF(subscription);
+    self->subscription = subscription;
+
     // determine type
     status = OCIAttrGet(descriptor, OCI_DTYPE_CHDES, &self->type, NULL,
             OCI_ATTR_CHDES_NFYTYPE, env->errorHandle);
@@ -414,47 +585,94 @@ static int Message_Initialize(
     if (Environment_CheckForError(env, status,
             "Message_Initialize(): get database name") < 0)
         return -1;
-    self->dbname = cxString_FromEncodedString(dbname, dbnameLength);
+    self->dbname = cxString_FromEncodedString(dbname, dbnameLength,
+            env->encoding);
     if (!self->dbname)
         return -1;
 
-    // determine table collection
-    status = OCIAttrGet(descriptor, OCI_DTYPE_CHDES, &tables, NULL,
-            OCI_ATTR_CHDES_TABLE_CHANGES, env->errorHandle);
-    if (Environment_CheckForError(env, status,
-            "Message_Initialize(): get tables collection") < 0)
-        return -1;
-
-    // determine number of tables
-    if (!tables)
-        numTables = 0;
-    else {
-        status = OCICollSize(env->handle, env->errorHandle, tables, &numTables);
+    if (self->type == OCI_EVENT_OBJCHANGE) {
+        // determine table collection
+        status = OCIAttrGet(descriptor, OCI_DTYPE_CHDES, &tables, NULL,
+                OCI_ATTR_CHDES_TABLE_CHANGES, env->errorHandle);
         if (Environment_CheckForError(env, status,
-                "Message_Initialize(): get size of collection") < 0)
+                "Message_Initialize(): get tables collection") < 0)
             return -1;
+
+        // determine number of tables
+        if (!tables)
+            numTables = 0;
+        else {
+            status = OCICollSize(env->handle, env->errorHandle, tables,
+                    &numTables);
+            if (Environment_CheckForError(env, status,
+                    "Message_Initialize(): get size of collection") < 0)
+                return -1;
+        }
+
+        // create list to hold results
+        self->tables = PyList_New(numTables);
+        if (!self->tables)
+            return -1;
+
+        // populate each entry with a message table instance
+        for (i = 0; i < numTables; i++) {
+            status = OCICollGetElem(env->handle, env->errorHandle, tables, i,
+                    &exists, (dvoid*) &tableDescriptor, &indicator);
+            if (Environment_CheckForError(env, status,
+                    "Message_Initialize(): get element from collection") < 0)
+                return -1;
+            table = (udt_MessageTable*)
+                    g_MessageTableType.tp_alloc(&g_MessageTableType, 0);
+            if (!table)
+                return -1;
+            PyList_SET_ITEM(self->tables, i, (PyObject*) table);
+            if (MessageTable_Initialize(table, env, *tableDescriptor) < 0)
+                return -1;
+        }
     }
 
-    // create list to hold results
-    self->tables = PyList_New(numTables);
-    if (!self->tables)
-        return -1;
-
-    // populate each entry with a message table instance
-    for (i = 0; i < numTables; i++) {
-        status = OCICollGetElem(env->handle, env->errorHandle, tables, i,
-                &exists, (dvoid*) &tableDescriptor, &indicator);
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+    if (self->type == OCI_EVENT_QUERYCHANGE) {
+        // determine query collection
+        status = OCIAttrGet(descriptor, OCI_DTYPE_CHDES, &queries, NULL,
+                OCI_ATTR_CHDES_QUERIES, env->errorHandle);
         if (Environment_CheckForError(env, status,
-                "Message_Initialize(): get element from collection") < 0)
+                "Message_Initialize(): get queries collection") < 0)
             return -1;
-        table = (udt_MessageTable*)
-                g_MessageTableType.tp_alloc(&g_MessageTableType, 0);
-        if (!table)
-            return -1;
-        PyList_SET_ITEM(self->tables, i, (PyObject*) table);
-        if (MessageTable_Initialize(table, env, *tableDescriptor) < 0)
+
+        // determine number of queries
+        if (!queries)
+            numQueries = 0;
+        else {
+            status = OCICollSize(env->handle, env->errorHandle, queries,
+                    &numQueries);
+            if (Environment_CheckForError(env, status,
+                    "Message_Initialize(): get size of collection") < 0)
+                return -1;
+        }
+
+        // create list to hold results
+        self->queries = PyList_New(numQueries);
+        if (!self->queries)
             return -1;
+
+        // populate each entry with a message query instance
+        for (i = 0; i < numQueries; i++) {
+            status = OCICollGetElem(env->handle, env->errorHandle, queries, i,
+                    &exists, (dvoid*) &queryDescriptor, &indicator);
+            if (Environment_CheckForError(env, status,
+                    "Message_Initialize(): get element from collection") < 0)
+                return -1;
+            query = (udt_MessageQuery*)
+                    g_MessageQueryType.tp_alloc(&g_MessageQueryType, 0);
+            if (!query)
+                return -1;
+            PyList_SET_ITEM(self->queries, i, (PyObject*) query);
+            if (MessageQuery_Initialize(query, env, *queryDescriptor) < 0)
+                return -1;
+        }
     }
+#endif
 
     return 0;
 }
@@ -476,7 +694,7 @@ static int Subscription_CallbackHandler(
     message = (udt_Message*) g_MessageType.tp_alloc(&g_MessageType, 0);
     if (!message)
         return -1;
-    if (Message_Initialize(message, env, descriptor) < 0) {
+    if (Message_Initialize(message, env, self, descriptor) < 0) {
         Py_DECREF(message);
         return -1;
     }
@@ -511,28 +729,12 @@ static void Subscription_Callback(
     ub4 mode)                           // mode used
 {
 #ifdef WITH_THREAD
-    PyThreadState *threadState;
+    PyGILState_STATE gstate = PyGILState_Ensure();
 #endif
     udt_Environment *env;
 
-    // create new thread state, if necessary
-#ifdef WITH_THREAD
-    threadState = PyThreadState_Swap(NULL);
-    if (threadState) {
-        PyThreadState_Swap(threadState);
-        threadState = NULL;
-    } else {
-        threadState = PyThreadState_New(g_InterpreterState);
-        if (!threadState) {
-            PyErr_Print();
-            return;
-        }
-        PyEval_AcquireThread(threadState);
-    }
-#endif
-
     // perform the call
-    env = Environment_NewFromScratch(0, 0);
+    env = Environment_NewFromScratch(0, 0, NULL, NULL);
     if (!env)
         PyErr_Print();
     else {
@@ -543,11 +745,7 @@ static void Subscription_Callback(
 
     // restore thread state, if necessary
 #ifdef WITH_THREAD
-    if (threadState) {
-        PyThreadState_Clear(threadState);
-        PyEval_ReleaseThread(threadState);
-        PyThreadState_Delete(threadState);
-    }
+    PyGILState_Release(gstate);
 #endif
 }
 
@@ -567,7 +765,7 @@ static int Subscription_Register(
     status = OCIHandleAlloc(env->handle, (dvoid**) &self->handle,
             OCI_HTYPE_SUBSCRIPTION, 0, 0);
     if (Environment_CheckForError(env, status,
-                "Subscription_Register(): allocate handle") < 0)
+            "Subscription_Register(): allocate handle") < 0)
         return -1;
 
     // set the namespace
@@ -575,7 +773,7 @@ static int Subscription_Register(
             (dvoid*) &self->namespace, sizeof(ub4), OCI_ATTR_SUBSCR_NAMESPACE,
             env->errorHandle);
     if (Environment_CheckForError(env, status,
-                "Subscription_Register(): set namespace") < 0)
+            "Subscription_Register(): set namespace") < 0)
         return -1;
 
     // set the protocol
@@ -583,7 +781,7 @@ static int Subscription_Register(
             (dvoid*) &self->protocol, sizeof(ub4), OCI_ATTR_SUBSCR_RECPTPROTO,
             env->errorHandle);
     if (Environment_CheckForError(env, status,
-                "Subscription_Register(): set protocol") < 0)
+            "Subscription_Register(): set protocol") < 0)
         return -1;
 
     // set the timeout
@@ -591,9 +789,19 @@ static int Subscription_Register(
             (dvoid*) &self->timeout, sizeof(ub4), OCI_ATTR_SUBSCR_TIMEOUT,
             env->errorHandle);
     if (Environment_CheckForError(env, status,
-                "Subscription_Register(): set timeout") < 0)
+            "Subscription_Register(): set timeout") < 0)
         return -1;
 
+    // set the TCP port used on client to listen for callback from DB server
+    if (self->port > 0) {
+        status = OCIAttrSet(env->handle, OCI_HTYPE_ENV,
+                (dvoid*) &(self->port), (ub4) 0, OCI_ATTR_SUBSCR_PORTNO,
+                env->errorHandle);
+        if (Environment_CheckForError(env, status,
+                "Subscription_Register(): set port") < 0)
+            return -1;
+    }
+
     // set the context for the callback
     status = OCIAttrSet(self->handle, OCI_HTYPE_SUBSCRIPTION,
             (dvoid*) self, 0, OCI_ATTR_SUBSCR_CTX, env->errorHandle);
@@ -611,6 +819,24 @@ static int Subscription_Register(
             return -1;
     }
 
+    // set suscription QOS
+    status = OCIAttrSet(self->handle, OCI_HTYPE_SUBSCRIPTION,
+            (dvoid*) &self->qos, sizeof(ub4), OCI_ATTR_SUBSCR_QOSFLAGS,
+            env->errorHandle);
+    if (Environment_CheckForError(env, status,
+                "Subscription_Register(): set qos flags") < 0)
+        return -1;
+
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+    // set subscription change notification QOS flags
+    status = OCIAttrSet(self->handle, OCI_HTYPE_SUBSCRIPTION,
+            (dvoid*) &self->cqqos, sizeof(ub4), OCI_ATTR_SUBSCR_CQ_QOSFLAGS,
+            env->errorHandle);
+    if (Environment_CheckForError(env, status,
+                "Subscription_Register(): set cq qos flags") < 0)
+        return -1;
+#endif
+
     // set whether or not rowids are desired
     status = OCIAttrSet(self->handle, OCI_HTYPE_SUBSCRIPTION,
             (dvoid*) &self->rowids, sizeof(ub4), OCI_ATTR_CHNF_ROWIDS,
@@ -636,6 +862,16 @@ static int Subscription_Register(
                 "Subscription_Register(): register") < 0)
         return -1;
 
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+    // get the registration id
+    status = OCIAttrGet(self->handle, OCI_HTYPE_SUBSCRIPTION, &self->id,
+              NULL, OCI_ATTR_SUBSCR_CQ_REGID, env->errorHandle);
+    if (Environment_CheckForError(env, status,
+                "Subscription_Register(): get registration id") < 0) {
+        return -1;
+    }
+#endif
+
     return 0;
 }
 
@@ -648,9 +884,12 @@ static udt_Subscription *Subscription_New(
     udt_Connection *connection,         // connection object
     ub4 namespace,                      // namespace to use
     ub4 protocol,                       // protocol to use
+    ub4 port,                           // client port for callbacks
     PyObject *callback,                 // callback routine
     ub4 timeout,                        // timeout (in seconds)
     ub4 operations,                     // operations to notify
+    ub4 qos,                            // QOS flags
+    ub4 cqqos,                          // change notification QOS flags
     int rowids)                         // retrieve rowids?
 {
     udt_Subscription *self;
@@ -665,10 +904,14 @@ static udt_Subscription *Subscription_New(
     self->callback = callback;
     self->namespace = namespace;
     self->protocol = protocol;
+    self->port = port;
     self->timeout = timeout;
     self->rowids = rowids;
     self->operations = operations;
+    self->qos = qos;
+    self->cqqos = cqqos;
     self->handle = NULL;
+    self->id = 0;
     if (Subscription_Register(self) < 0) {
         Py_DECREF(self);
         return NULL;
@@ -741,9 +984,12 @@ static PyObject *Subscription_RegisterQuery(
     PyObject *args)                     // arguments
 {
     PyObject *statement, *executeArgs;
-    udt_StringBuffer statementBuffer;
+    udt_Buffer statementBuffer;
     udt_Environment *env;
     udt_Cursor *cursor;
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+    ub8 queryid;
+#endif
     sword status;
 
     // parse arguments
@@ -772,14 +1018,15 @@ static PyObject *Subscription_RegisterQuery(
     }
 
     // prepare the statement for execution
-    if (StringBuffer_Fill(&statementBuffer, statement) < 0) {
+    if (cxBuffer_FromObject(&statementBuffer, statement,
+            env->encoding) < 0) {
         Py_DECREF(cursor);
         return NULL;
     }
     status = OCIStmtPrepare(cursor->handle, env->errorHandle,
             (text*) statementBuffer.ptr, statementBuffer.size, OCI_NTV_SYNTAX,
             OCI_DEFAULT);
-    StringBuffer_Clear(&statementBuffer);
+    cxBuffer_Clear(&statementBuffer);
     if (Environment_CheckForError(env, status,
             "Subscription_RegisterQuery(): prepare statement") < 0) {
         Py_DECREF(cursor);
@@ -828,8 +1075,27 @@ static PyObject *Subscription_RegisterQuery(
         Py_DECREF(cursor);
         return NULL;
     }
+
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+    if (self->cqqos & OCI_SUBSCR_CQ_QOS_QUERY) {
+        // get the query id
+        status = OCIAttrGet(cursor->handle, OCI_HTYPE_STMT, &queryid, NULL,
+                OCI_ATTR_CQ_QUERYID, env->errorHandle);
+        if (Environment_CheckForError(env, status,
+                    "Subscription_RegisterQuery(): get query id") < 0) {
+            Py_DECREF(cursor);
+            return NULL;
+        }
+    }
+#endif
+
     Py_DECREF(cursor);
 
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+    if (self->cqqos & OCI_SUBSCR_CQ_QOS_QUERY)
+        return PyInt_FromLong(queryid);
+#endif
+
     Py_INCREF(Py_None);
     return Py_None;
 }
@@ -842,8 +1108,10 @@ static PyObject *Subscription_RegisterQuery(
 static void Message_Free(
     udt_Message *self)                  // object to free
 {
+    Py_CLEAR(self->subscription);
     Py_CLEAR(self->dbname);
     Py_CLEAR(self->tables);
+    Py_CLEAR(self->queries);
     Py_TYPE(self)->tp_free((PyObject*) self);
 }
 
@@ -856,6 +1124,7 @@ static void MessageTable_Free(
     udt_MessageTable *self)             // object to free
 {
     Py_CLEAR(self->name);
+    Py_CLEAR(self->rows);
     Py_TYPE(self)->tp_free((PyObject*) self);
 }
 
@@ -871,3 +1140,17 @@ static void MessageRow_Free(
     Py_TYPE(self)->tp_free((PyObject*) self);
 }
 
+
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
+//-----------------------------------------------------------------------------
+// MessageQuery_Free()
+//   Free the memory associated with a query in a message.
+//-----------------------------------------------------------------------------
+static void MessageQuery_Free(
+    udt_MessageQuery *self)               // object to free
+{
+    Py_CLEAR(self->tables);
+    Py_TYPE(self)->tp_free((PyObject*) self);
+}
+#endif
+
