@@ -41,6 +41,7 @@ typedef int (*InitializeProc)(udt_Variable*, udt_Cursor*);
 typedef void (*FinalizeProc)(udt_Variable*);
 typedef int (*PreDefineProc)(udt_Variable*, OCIParam*);
 typedef int (*PostDefineProc)(udt_Variable*);
+typedef int (*PreFetchProc)(udt_Variable*);
 typedef int (*IsNullProc)(udt_Variable*, unsigned);
 typedef int (*SetValueProc)(udt_Variable*, unsigned, PyObject*);
 typedef PyObject * (*GetValueProc)(udt_Variable*, unsigned);
@@ -55,6 +56,7 @@ typedef struct _udt_VariableType {
     FinalizeProc finalizeProc;
     PreDefineProc preDefineProc;
     PostDefineProc postDefineProc;
+    PreFetchProc preFetchProc;
     IsNullProc isNullProc;
     SetValueProc setValueProc;
     GetValueProc getValueProc;
@@ -361,9 +363,10 @@ static int Variable_Check(
             Py_TYPE(object) == &g_StringVarType ||
             Py_TYPE(object) == &g_FixedCharVarType ||
             Py_TYPE(object) == &g_NCLOBVarType ||
-#ifndef WITH_UNICODE
+#if PY_MAJOR_VERSION < 3
             Py_TYPE(object) == &g_UnicodeVarType ||
             Py_TYPE(object) == &g_FixedUnicodeVarType ||
+            Py_TYPE(object) == &g_LongUnicodeVarType ||
 #endif
             Py_TYPE(object) == &g_RowidVarType ||
             Py_TYPE(object) == &g_BinaryVarType ||
@@ -391,13 +394,15 @@ static udt_VariableType *Variable_TypeByPythonType(
         return &vt_String;
     if (type == (PyObject*) &g_FixedCharVarType)
         return &vt_FixedChar;
-#ifndef WITH_UNICODE
+#if PY_MAJOR_VERSION < 3
     if (type == (PyObject*) &g_UnicodeVarType)
         return &vt_NationalCharString;
     if (type == (PyObject*) &PyUnicode_Type)
         return &vt_NationalCharString;
     if (type == (PyObject*) &g_FixedUnicodeVarType)
         return &vt_FixedNationalChar;
+    if (type == (PyObject*) &g_LongUnicodeVarType)
+        return &vt_LongNationalCharString;
 #endif
     if (type == (PyObject*) &g_NCLOBVarType)
         return &vt_NCLOB;
@@ -450,6 +455,8 @@ static udt_VariableType *Variable_TypeByPythonType(
     if (type == (PyObject*) &g_NativeFloatVarType)
         return &vt_NativeFloat;
 #endif
+    if (type == (PyObject*) &g_ObjectVarType)
+        return &vt_Object;
 
     PyErr_SetString(g_NotSupportedErrorException,
             "Variable_TypeByPythonType(): unhandled data type");
@@ -483,33 +490,33 @@ static udt_VariableType *Variable_TypeByValue(
             return &vt_LongString;
         return &vt_String;
     }
-#ifdef WITH_UNICODE
+#if PY_MAJOR_VERSION < 3
+    if (PyUnicode_Check(value)) {
+        *size = PyUnicode_GET_SIZE(value);
+        if (*size > MAX_STRING_CHARS)
+            return &vt_LongNationalCharString;
+        return &vt_NationalCharString;
+    }
+    if (PyInt_Check(value))
+        return &vt_Integer;
+#else
     if (PyBytes_Check(value)) {
         *size = PyBytes_GET_SIZE(value);
         if (*size > MAX_BINARY_BYTES)
             return &vt_LongBinary;
         return &vt_Binary;
     }
-#else
-    if (PyUnicode_Check(value)) {
-        *size = PyUnicode_GET_SIZE(value);
-        return &vt_NationalCharString;
-    }
-#endif
-#if PY_MAJOR_VERSION < 3
-    if (PyInt_Check(value))
-        return &vt_Integer;
 #endif
     if (PyLong_Check(value))
         return &vt_LongInteger;
     if (PyFloat_Check(value))
         return &vt_Float;
     if (cxBinary_Check(value)) {
-        udt_StringBuffer temp;
-        if (StringBuffer_FromBinary(&temp, value) < 0)
+        udt_Buffer temp;
+        if (cxBuffer_FromObject(&temp, value, NULL) < 0)
             return NULL;
         *size = temp.size;
-        StringBuffer_Clear(&temp);
+        cxBuffer_Clear(&temp);
         if (*size > MAX_BINARY_BYTES)
             return &vt_LongBinary;
         return &vt_Binary;
@@ -570,16 +577,12 @@ static udt_VariableType *Variable_TypeByOracleDataType (
         case SQLT_LNG:
             return &vt_LongString;
         case SQLT_AFC:
-#ifndef WITH_UNICODE
             if (charsetForm == SQLCS_NCHAR)
                 return &vt_FixedNationalChar;
-#endif
             return &vt_FixedChar;
         case SQLT_CHR:
-#ifndef WITH_UNICODE
             if (charsetForm == SQLCS_NCHAR)
                 return &vt_NationalCharString;
-#endif
             return &vt_String;
         case SQLT_RDD:
             return &vt_Rowid;
@@ -875,7 +878,7 @@ static udt_Variable *Variable_NewByOutputTypeHandler(
         return NULL;
 
     // retrieve scale and precision of the parameter, if applicable
-    scale = precision = 0;
+    precision = scale = 0;
     if (varType->pythonType == &g_NumberVarType) {
         status = OCIAttrGet(param, OCI_HTYPE_DESCRIBE, (dvoid*) &scale, 0,
                 OCI_ATTR_SCALE, cursor->environment->errorHandle);
@@ -1052,8 +1055,9 @@ static int Variable_InternalBind(
 
     // perform the bind
     if (var->boundName) {
-        udt_StringBuffer buffer;
-        if (StringBuffer_Fill(&buffer, var->boundName) < 0)
+        udt_Buffer buffer;
+        if (cxBuffer_FromObject(&buffer, var->boundName,
+                var->environment->encoding) < 0)
             return -1;
         if (var->isArray) {
             status = OCIBindByName(var->boundCursorHandle, &var->bindHandle,
@@ -1069,7 +1073,7 @@ static int Variable_InternalBind(
                     var->type->oracleType, var->indicator, var->actualLength,
                     var->returnCode, 0, 0, OCI_DEFAULT);
         }
-        StringBuffer_Clear(&buffer);
+        cxBuffer_Clear(&buffer);
     } else {
         if (var->isArray) {
             status = OCIBindByPos(var->boundCursorHandle, &var->bindHandle,
@@ -1088,22 +1092,14 @@ static int Variable_InternalBind(
             "Variable_InternalBind()") < 0)
         return -1;
 
-#ifndef WITH_UNICODE
     // set the charset form and id if applicable
     if (var->type->charsetForm != SQLCS_IMPLICIT) {
-        ub2 charsetId = OCI_UTF16ID;
         status = OCIAttrSet(var->bindHandle, OCI_HTYPE_BIND,
                 (dvoid*) &var->type->charsetForm, 0, OCI_ATTR_CHARSET_FORM,
                 var->environment->errorHandle);
         if (Environment_CheckForError(var->environment, status,
                 "Variable_InternalBind(): set charset form") < 0)
             return -1;
-        status = OCIAttrSet(var->bindHandle, OCI_HTYPE_BIND,
-                 (dvoid*) &charsetId, 0, OCI_ATTR_CHARSET_ID,
-                 var->environment->errorHandle);
-        if (Environment_CheckForError(var->environment, status,
-                 "Variable_InternalBind(): setting charset Id") < 0)
-            return -1;
         status = OCIAttrSet(var->bindHandle, OCI_HTYPE_BIND,
                 (dvoid*) &var->bufferSize, 0, OCI_ATTR_MAXDATA_SIZE,
                 var->environment->errorHandle);
@@ -1111,7 +1107,6 @@ static int Variable_InternalBind(
                 "Variable_InternalBind(): set max data size") < 0)
             return -1;
     }
-#endif
 
     // set the max data size for strings
     if ((var->type == &vt_String || var->type == &vt_FixedChar)
@@ -1146,14 +1141,6 @@ static int Variable_Bind(
     var->boundPos = pos;
     var->boundCursorHandle = cursor->handle;
     Py_XDECREF(var->boundName);
-#if defined(WITH_UNICODE) && PY_MAJOR_VERSION < 3
-    if (name && PyBytes_Check(name)) {
-        var->boundName = PyObject_Unicode(name);
-        if (!var->boundName)
-            return -1;
-        return Variable_InternalBind(var);
-    }
-#endif
     Py_XINCREF(name);
     var->boundName = name;
 
@@ -1364,8 +1351,14 @@ static int Variable_SetValue(
     unsigned arrayPos,                  // array position
     PyObject *value)                    // value to set
 {
-    if (var->isArray)
+    if (var->isArray) {
+        if (arrayPos > 0) {
+            PyErr_SetString(g_NotSupportedErrorException,
+                    "arrays of arrays are not supported by the OCI");
+            return -1;
+        }
         return Variable_SetArrayValue(var, value);
+    }
     return Variable_SetSingleValue(var, arrayPos, value);
 }
 
