@@ -11,15 +11,20 @@
 #include <orid.h>
 #include <xa.h>
 
-// define whether or not we are building Oracle 9i
-#ifdef OCI_ATTR_MODULE
-#define ORACLE_10G
+// define simple way to respresent Oracle version
+#define ORACLE_VERSION(major, minor) \
+        ((major << 8) | minor)
+
+// define what version of Oracle we are building as 2 byte hex number
+#if !defined(OCI_MAJOR_VERSION) && defined(OCI_ATTR_MODULE)
+#define OCI_MAJOR_VERSION 10
+#define OCI_MINOR_VERSION 1
 #endif
-#ifdef OCI_MAJOR_VERSION
-#define ORACLE_10GR2
-#endif
-#ifdef OCI_ATTR_CONNECTION_CLASS
-#define ORACLE_11G
+#if defined(OCI_MAJOR_VERSION) && defined(OCI_MINOR_VERSION)
+#define ORACLE_VERSION_HEX \
+        ORACLE_VERSION(OCI_MAJOR_VERSION, OCI_MINOR_VERSION)
+#else
+#error Unsupported version of OCI.
 #endif
 
 // PY_LONG_LONG was called LONG_LONG before Python 2.3
@@ -58,17 +63,57 @@ typedef int Py_ssize_t;
 #define PyInt_Type              PyLong_Type
 #endif
 
+// use the bytes methods in cx_Oracle and define them as the equivalent string
+// type methods as is done in Python 2.6
+#ifndef PyBytes_Check
+    #define PyBytes_Type                PyString_Type
+    #define PyBytes_AS_STRING           PyString_AS_STRING
+    #define PyBytes_GET_SIZE            PyString_GET_SIZE
+    #define PyBytes_Check               PyString_Check
+    #define PyBytes_Format              PyString_Format
+    #define PyBytes_FromString          PyString_FromString
+    #define PyBytes_FromStringAndSize   PyString_FromStringAndSize
+#endif
+
+// define types and methods for strings and binary data
+#if PY_MAJOR_VERSION >= 3
+    #define cxBinary_Type               PyBytes_Type
+    #define cxBinary_Check              PyBytes_Check
+    #define cxString_Type               &PyUnicode_Type
+    #define cxString_Format             PyUnicode_Format
+    #define cxString_Check              PyUnicode_Check
+    #define cxString_GetSize            PyUnicode_GET_SIZE
+#else
+    #define cxBinary_Type               PyBuffer_Type
+    #define cxBinary_Check              PyBuffer_Check
+    #define cxString_Type               &PyBytes_Type
+    #define cxString_Format             PyBytes_Format
+    #define cxString_Check              PyBytes_Check
+    #define cxString_GetSize            PyBytes_GET_SIZE
+#endif
+
+// define string type and methods
+#if PY_MAJOR_VERSION >= 3
+    #define cxString_FromAscii(str) \
+        PyUnicode_DecodeASCII(str, strlen(str), NULL)
+    #define cxString_FromEncodedString(buffer, numBytes, encoding) \
+        PyUnicode_Decode(buffer, numBytes, encoding, NULL)
+#else
+    #define cxString_FromAscii(str) \
+        PyBytes_FromString(str)
+    #define cxString_FromEncodedString(buffer, numBytes, encoding) \
+        PyBytes_FromStringAndSize(buffer, numBytes)
+#endif
+
 // define base exception
 #if PY_MAJOR_VERSION >= 3
 #define CXORA_BASE_EXCEPTION    NULL
+#define CXORA_TYPE_ERROR        "expecting string or bytes object"
 #else
 #define CXORA_BASE_EXCEPTION    PyExc_StandardError
+#define CXORA_TYPE_ERROR        "expecting string, unicode or buffer object"
 #endif
 
-// define simple construct for determining endianness of the platform
-// Oracle uses native encoding with OCI_UTF16 but bails when a BOM is written
-#define IS_LITTLE_ENDIAN (int)*(unsigned char*) &one
-
 // define macro for adding OCI constants
 #define ADD_OCI_CONSTANT(x) \
     if (PyModule_AddIntConstant(module, #x, OCI_ ##x) < 0) \
@@ -94,14 +139,11 @@ typedef int Py_ssize_t;
 #define BUILD_VERSION_STRING    xstr(BUILD_VERSION)
 #define DRIVER_NAME             "cx_Oracle-"BUILD_VERSION_STRING
 
-#include "StringUtils.c"
+#include "Buffer.c"
 
 //-----------------------------------------------------------------------------
 // Globals
 //-----------------------------------------------------------------------------
-#ifdef WITH_THREAD
-static PyInterpreterState *g_InterpreterState;
-#endif
 static PyObject *g_WarningException = NULL;
 static PyObject *g_ErrorException = NULL;
 static PyObject *g_InterfaceErrorException = NULL;
@@ -114,10 +156,6 @@ static PyObject *g_ProgrammingErrorException = NULL;
 static PyObject *g_NotSupportedErrorException = NULL;
 static PyTypeObject *g_DateTimeType = NULL;
 static PyTypeObject *g_DecimalType = NULL;
-static PyObject *g_ShortNumberToStringFormatObj = NULL;
-static udt_StringBuffer g_ShortNumberToStringFormatBuffer;
-static PyObject *g_NumberToStringFormatObj = NULL;
-static udt_StringBuffer g_NumberToStringFormatBuffer;
 
 
 //-----------------------------------------------------------------------------
@@ -171,21 +209,44 @@ static int GetModuleAndName(
 //-----------------------------------------------------------------------------
 static PyObject* MakeDSN(
     PyObject* self,                     // passthrough argument
-    PyObject* args)                     // arguments to function
+    PyObject* args,                     // arguments to function
+    PyObject* keywordArgs)              // keyword arguments
 {
-    PyObject *format, *result;
+    static char *keywordList[] = { "host", "port", "sid", "service_name",
+            NULL };
+    PyObject *hostObj, *portObj, *sidObj, *serviceNameObj, *connectDataObj;
+    PyObject *format, *result, *formatArgs;
 
-    format = cxString_FromAscii("(DESCRIPTION=(ADDRESS_LIST=(ADDRESS="
-            "(PROTOCOL=TCP)(HOST=%s)(PORT=%s)))(CONNECT_DATA=(SID=%s)))");
+    // parse arguments
+    sidObj = serviceNameObj = NULL;
+    if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "OO|OO", keywordList,
+            &hostObj, &portObj, &sidObj, &serviceNameObj))
+        return NULL;
+    if (sidObj) {
+        connectDataObj = sidObj;
+        format = cxString_FromAscii("(DESCRIPTION=(ADDRESS_LIST=(ADDRESS="
+                "(PROTOCOL=TCP)(HOST=%s)(PORT=%s)))(CONNECT_DATA=(SID=%s)))");
+    } else {
+        connectDataObj = serviceNameObj;
+        format = cxString_FromAscii("(DESCRIPTION=(ADDRESS_LIST=(ADDRESS="
+                "(PROTOCOL=TCP)(HOST=%s)(PORT=%s)))(CONNECT_DATA="
+                "(SERVICE_NAME=%s)))");
+    }
     if (!format)
         return NULL;
-    result = cxString_Format(format, args);
+    formatArgs = PyTuple_Pack(3, hostObj, portObj, connectDataObj);
+    if (!formatArgs) {
+        Py_DECREF(format);
+        return NULL;
+    }
+    result = cxString_Format(format, formatArgs);
     Py_DECREF(format);
+    Py_DECREF(formatArgs);
     return result;
 }
 
 
-#ifdef ORACLE_10GR2
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
 //-----------------------------------------------------------------------------
 // ClientVersion()
 //   Return the version of the Oracle client being used as a 5-tuple.
@@ -260,13 +321,13 @@ static PyObject* TimestampFromTicks(
 //   Declaration of methods supported by this module
 //-----------------------------------------------------------------------------
 static PyMethodDef g_ModuleMethods[] = {
-    { "makedsn", MakeDSN, METH_VARARGS },
-    { "Time", Time, METH_VARARGS },
-    { "DateFromTicks", DateFromTicks, METH_VARARGS },
-    { "TimeFromTicks", TimeFromTicks, METH_VARARGS },
-    { "TimestampFromTicks", TimestampFromTicks, METH_VARARGS },
-#ifdef ORACLE_10GR2
-    { "clientversion", ClientVersion, METH_NOARGS },
+    { "makedsn", (PyCFunction) MakeDSN, METH_VARARGS | METH_KEYWORDS },
+    { "Time", (PyCFunction) Time, METH_VARARGS },
+    { "DateFromTicks", (PyCFunction) DateFromTicks, METH_VARARGS },
+    { "TimeFromTicks", (PyCFunction) TimeFromTicks, METH_VARARGS },
+    { "TimestampFromTicks", (PyCFunction) TimestampFromTicks, METH_VARARGS },
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
+    { "clientversion", (PyCFunction) ClientVersion, METH_NOARGS },
 #endif
     { NULL }
 };
@@ -296,17 +357,10 @@ static struct PyModuleDef g_ModuleDef = {
 //-----------------------------------------------------------------------------
 static PyObject *Module_Initialize(void)
 {
-    PyThreadState *threadState;
     PyObject *module;
 
-    // initialize the interpreter state for callbacks
 #ifdef WITH_THREAD
     PyEval_InitThreads();
-    threadState = PyThreadState_Swap(NULL);
-    if (!threadState)
-        return NULL;
-    g_InterpreterState = threadState->interp;
-    PyThreadState_Swap(threadState);
 #endif
 
     // import the datetime module for datetime support
@@ -322,21 +376,6 @@ static PyObject *Module_Initialize(void)
     if (!g_DecimalType)
         return NULL;
 
-    // set up the string and buffer for converting numbers to strings
-    g_ShortNumberToStringFormatObj = cxString_FromAscii("TM9");
-    if (!g_ShortNumberToStringFormatObj)
-        return NULL;
-    if (StringBuffer_Fill(&g_ShortNumberToStringFormatBuffer,
-            g_ShortNumberToStringFormatObj) < 0)
-        return NULL;
-    g_NumberToStringFormatObj = cxString_FromAscii(
-            "999999999999999999999999999999999999999999999999999999999999999");
-    if (!g_NumberToStringFormatObj)
-        return NULL;
-    if (StringBuffer_Fill(&g_NumberToStringFormatBuffer,
-            g_NumberToStringFormatObj) < 0)
-        return NULL;
-
     // prepare the types for use by the module
     MAKE_TYPE_READY(&g_ConnectionType);
     MAKE_TYPE_READY(&g_CursorType);
@@ -347,11 +386,14 @@ static PyObject *Module_Initialize(void)
     MAKE_TYPE_READY(&g_ObjectAttributeType);
     MAKE_TYPE_READY(&g_ExternalLobVarType);
     MAKE_TYPE_READY(&g_ExternalObjectVarType);
-#ifdef ORACLE_10GR2
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
     MAKE_TYPE_READY(&g_SubscriptionType);
     MAKE_TYPE_READY(&g_MessageType);
     MAKE_TYPE_READY(&g_MessageTableType);
     MAKE_TYPE_READY(&g_MessageRowType);
+#endif
+#if ORACLE_VERSION_HEX > ORACLE_VERSION(11, 1)
+    MAKE_TYPE_READY(&g_MessageQueryType);
 #endif
     MAKE_VARIABLE_TYPE_READY(&g_StringVarType);
     MAKE_VARIABLE_TYPE_READY(&g_FixedCharVarType);
@@ -367,9 +409,10 @@ static PyObject *Module_Initialize(void)
     MAKE_VARIABLE_TYPE_READY(&g_BFILEVarType);
     MAKE_VARIABLE_TYPE_READY(&g_CursorVarType);
     MAKE_VARIABLE_TYPE_READY(&g_ObjectVarType);
-#ifndef WITH_UNICODE
+#if PY_MAJOR_VERSION < 3
     MAKE_VARIABLE_TYPE_READY(&g_UnicodeVarType);
     MAKE_VARIABLE_TYPE_READY(&g_FixedUnicodeVarType);
+    MAKE_VARIABLE_TYPE_READY(&g_LongUnicodeVarType);
 #endif
     MAKE_VARIABLE_TYPE_READY(&g_NCLOBVarType);
 #ifdef SQLT_BFLOAT
@@ -439,9 +482,10 @@ static PyObject *Module_Initialize(void)
     ADD_TYPE_OBJECT("OBJECT", &g_ObjectVarType)
     ADD_TYPE_OBJECT("DATETIME", &g_DateTimeVarType)
     ADD_TYPE_OBJECT("FIXED_CHAR", &g_FixedCharVarType)
-#ifndef WITH_UNICODE
+#if PY_MAJOR_VERSION < 3
     ADD_TYPE_OBJECT("FIXED_UNICODE", &g_FixedUnicodeVarType)
     ADD_TYPE_OBJECT("UNICODE", &g_UnicodeVarType)
+    ADD_TYPE_OBJECT("LONG_UNICODE", &g_LongUnicodeVarType)
 #endif
     ADD_TYPE_OBJECT("INTERVAL", &g_IntervalVarType)
     ADD_TYPE_OBJECT("LOB", &g_ExternalLobVarType)
@@ -487,7 +531,7 @@ static PyObject *Module_Initialize(void)
     ADD_OCI_CONSTANT(SPOOL_ATTRVAL_WAIT)
     ADD_OCI_CONSTANT(SPOOL_ATTRVAL_NOWAIT)
     ADD_OCI_CONSTANT(SPOOL_ATTRVAL_FORCEGET)
-#ifdef ORACLE_10GR2
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(10, 2)
     ADD_OCI_CONSTANT(PRELIM_AUTH)
     ADD_OCI_CONSTANT(DBSHUTDOWN_ABORT)
     ADD_OCI_CONSTANT(DBSHUTDOWN_FINAL)
@@ -512,11 +556,24 @@ static PyObject *Module_Initialize(void)
     ADD_OCI_CONSTANT(SUBSCR_PROTO_MAIL)
     ADD_OCI_CONSTANT(SUBSCR_PROTO_SERVER)
     ADD_OCI_CONSTANT(SUBSCR_PROTO_HTTP)
+    ADD_OCI_CONSTANT(SUBSCR_QOS_RELIABLE)
+    ADD_OCI_CONSTANT(SUBSCR_QOS_PAYLOAD)
+    ADD_OCI_CONSTANT(SUBSCR_QOS_REPLICATE)
+    ADD_OCI_CONSTANT(SUBSCR_QOS_SECURE)
+    ADD_OCI_CONSTANT(SUBSCR_QOS_PURGE_ON_NTFN)
+    ADD_OCI_CONSTANT(SUBSCR_QOS_MULTICBK)
 #endif
-#ifdef ORACLE_11G
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 1)
     ADD_OCI_CONSTANT(ATTR_PURITY_DEFAULT)
     ADD_OCI_CONSTANT(ATTR_PURITY_NEW)
     ADD_OCI_CONSTANT(ATTR_PURITY_SELF)
+    ADD_OCI_CONSTANT(EVENT_QUERYCHANGE)
+    ADD_OCI_CONSTANT(SUBSCR_CQ_QOS_QUERY)
+    ADD_OCI_CONSTANT(SUBSCR_CQ_QOS_BEST_EFFORT)
+    ADD_OCI_CONSTANT(SUBSCR_CQ_QOS_CLQRYCACHE)
+#endif
+#if ORACLE_VERSION_HEX >= ORACLE_VERSION(11, 2)
+    ADD_OCI_CONSTANT(SUBSCR_QOS_HAREG)
 #endif
 
     return module;
